# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import logging
import os
import re

from codecs import BOM_UTF8
from typing import Any, Dict, Generator, Mapping, Sequence, Tuple, Optional

import chardet
import iso8601

from swh.model.identifiers import normalize_timestamp
from swh.loader.package.loader import PackageLoader
from swh.loader.package.utils import api_info, release_name


logger = logging.getLogger(__name__)


_EMPTY_AUTHOR = {'fullname': b'', 'name': None, 'email': None}

# https://github.com/jonschlinkert/author-regex
_author_regexp = r'([^<(]+?)?[ \t]*(?:<([^>(]+?)>)?[ \t]*(?:\(([^)]+?)\)|$)'


class NpmLoader(PackageLoader):
    visit_type = 'npm'

    def __init__(self, package_name, package_url, package_metadata_url):
        super().__init__(url=package_url)
        self.provider_url = package_metadata_url

        self._info = None
        self._versions = None

        # if package_url is None:
        #     package_url = 'https://www.npmjs.com/package/%s' % package_name
        # if package_metadata_url is None:
        #     package_metadata_url = 'https://replicate.npmjs.com/%s/' %\
        #                             quote(package_name, safe='')

    @property
    def info(self) -> Dict:
        """Return the project metadata information (fetched from npm registry)

        """
        if not self._info:
            self._info = api_info(self.provider_url)
        return self._info

    def get_versions(self) -> Sequence[str]:
        return sorted(list(self.info['versions'].keys()))

    def get_default_version(self) -> str:
        return self.info['dist-tags'].get('latest', '')

    def get_package_info(self, version: str) -> Generator[
            Tuple[str, Mapping[str, Any]], None, None]:
        meta = self.info['versions'][version]
        url = meta['dist']['tarball']
        p_info = {
            'url': url,
            'filename': os.path.basename(url),
            'raw': meta,
        }
        yield release_name(version), p_info

    def resolve_revision_from(
            self, known_artifacts: Dict, artifact_metadata: Dict) \
            -> Optional[bytes]:
        shasum = artifact_metadata['dist']['shasum']
        for rev_id, known_artifact in known_artifacts.items():
            original_artifact = known_artifact['original_artifact'][0]
            if shasum == original_artifact['checksums']['sha1']:
                return rev_id
        return None

    def build_revision(
            self, a_metadata: Dict, uncompressed_path: str) -> Dict:
        i_metadata = extract_intrinsic_metadata(uncompressed_path)
        # from intrinsic metadata
        author = extract_npm_package_author(i_metadata)
        # extrinsic metadata
        version = i_metadata['version']
        date = self.info['time'][version]
        date = iso8601.parse_date(date)
        date = normalize_timestamp(int(date.timestamp()))
        message = version.encode('ascii')

        return {
            'type': 'tar',
            'message': message,
            'author': author,
            'date': date,
            'committer': author,
            'committer_date': date,
            'parents': [],
            'metadata': {
                'intrinsic': {
                    'tool': 'package.json',
                    'raw': i_metadata,
                },
                'extrinsic': {
                    'provider': self.provider_url,
                    'when': self.visit_date.isoformat(),
                    'raw': a_metadata,
                },
            },
        }


def parse_npm_package_author(author_str):
    """
    Parse npm package author string.

    It works with a flexible range of formats, as detailed below::

        name
        name <email> (url)
        name <email>(url)
        name<email> (url)
        name<email>(url)
        name (url) <email>
        name (url)<email>
        name(url) <email>
        name(url)<email>
        name (url)
        name(url)
        name <email>
        name<email>
        <email> (url)
        <email>(url)
        (url) <email>
        (url)<email>
        <email>
        (url)

    Args:
        author_str (str): input author string

    Returns:
        dict: A dict that may contain the following keys:
            * name
            * email
            * url

    """
    author = {}
    matches = re.findall(_author_regexp,
                         author_str.replace('<>', '').replace('()', ''),
                         re.M)
    for match in matches:
        if match[0].strip():
            author['name'] = match[0].strip()
        if match[1].strip():
            author['email'] = match[1].strip()
        if match[2].strip():
            author['url'] = match[2].strip()
    return author


def extract_npm_package_author(package_json):
    """
    Extract package author from a ``package.json`` file content and
    return it in swh format.

    Args:
        package_json (dict): Dict holding the content of parsed
            ``package.json`` file

    Returns:
        dict: A dict with the following keys:
            * fullname
            * name
            * email

    """

    def _author_str(author_data):
        if type(author_data) is dict:
            author_str = ''
            if 'name' in author_data:
                author_str += author_data['name']
            if 'email' in author_data:
                author_str += ' <%s>' % author_data['email']
            return author_str
        elif type(author_data) is list:
            return _author_str(author_data[0]) if len(author_data) > 0 else ''
        else:
            return author_data

    author_data = {}
    for author_key in ('author', 'authors'):
        if author_key in package_json:
            author_str = _author_str(package_json[author_key])
            author_data = parse_npm_package_author(author_str)

    name = author_data.get('name')
    email = author_data.get('email')

    fullname = None

    if name and email:
        fullname = '%s <%s>' % (name, email)
    elif name:
        fullname = name

    if not fullname:
        return _EMPTY_AUTHOR

    if fullname:
        fullname = fullname.encode('utf-8')

    if name:
        name = name.encode('utf-8')

    if email:
        email = email.encode('utf-8')

    return {'fullname': fullname, 'name': name, 'email': email}


def _lstrip_bom(s, bom=BOM_UTF8):
    if s.startswith(bom):
        return s[len(bom):]
    else:
        return s


def load_json(json_bytes):
    """
    Try to load JSON from bytes and return a dictionary.

    First try to decode from utf-8. If the decoding failed,
    try to detect the encoding and decode again with replace
    error handling.

    If JSON is malformed, an empty dictionary will be returned.

    Args:
        json_bytes (bytes): binary content of a JSON file

    Returns:
        dict: JSON data loaded in a dictionary
    """
    json_data = {}
    try:
        json_str = _lstrip_bom(json_bytes).decode('utf-8')
    except UnicodeDecodeError:
        encoding = chardet.detect(json_bytes)['encoding']
        if encoding:
            json_str = json_bytes.decode(encoding, 'replace')
    try:
        json_data = json.loads(json_str)
    except json.decoder.JSONDecodeError:
        pass
    return json_data


def extract_intrinsic_metadata(dir_path: str) -> Dict:
    """Given an uncompressed path holding the pkginfo file, returns a
       pkginfo parsed structure as a dict.

       The release artifact contains at their root one folder. For example:
       $ tar tvf zprint-0.0.6.tar.gz
       drwxr-xr-x root/root         0 2018-08-22 11:01 zprint-0.0.6/
       ...

    Args:

        dir_path (str): Path to the uncompressed directory
                        representing a release artifact from npm.

    Returns:
        the pkginfo parsed structure as a dict if any or None if
        none was present.

    """
    # Retrieve the root folder of the archive
    if not os.path.exists(dir_path):
        return {}
    lst = os.listdir(dir_path)
    if len(lst) == 0:
        return {}
    project_dirname = lst[0]
    package_json_path = os.path.join(dir_path, project_dirname, 'package.json')
    if not os.path.exists(package_json_path):
        return {}
    with open(package_json_path, 'rb') as package_json_file:
        package_json_bytes = package_json_file.read()
        return load_json(package_json_bytes)



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import os
import pytest


import swh.loader.package
from swh.loader.package.utils import download, api_info, release_name


def test_version_generation():
    assert swh.loader.package.__version__ != 'devel', \
        "Make sure swh.loader.core is installed (e.g. pip install -e .)"


@pytest.mark.fs
def test_download_fail_to_download(tmp_path, requests_mock):
    url = 'https://pypi.org/pypi/arrow/json'
    status_code = 404
    requests_mock.get(url, status_code=status_code)

    with pytest.raises(ValueError) as e:
        download(url, tmp_path)

    assert e.value.args[0] == "Fail to query '%s'. Reason: %s" % (
        url, status_code)


@pytest.mark.fs
def test_download_fail_length_mismatch(tmp_path, requests_mock):
    """Mismatch length after download should raise

    """
    filename = 'requests-0.0.1.tar.gz'
    url = 'https://pypi.org/pypi/requests/%s' % filename
    data = 'this is something'
    wrong_size = len(data) - 3
    requests_mock.get(url, text=data, headers={
        'content-length': str(wrong_size)  # wrong size!
    })

    with pytest.raises(ValueError) as e:
        download(url, dest=str(tmp_path))

    assert e.value.args[0] == "Error when checking size: %s != %s" % (
        wrong_size, len(data)
    )


@pytest.mark.fs
def test_download_ok(tmp_path, requests_mock):
    """Download without issue should provide filename and hashes"""
    filename = 'requests-0.0.1.tar.gz'
    url = 'https://pypi.org/pypi/requests/%s' % filename
    data = 'this is something'
    requests_mock.get(url, text=data, headers={
        'content-length': str(len(data))
    })

    actual_filepath, actual_hashes = download(url, dest=str(tmp_path))

    actual_filename = os.path.basename(actual_filepath)
    assert actual_filename == filename
    assert actual_hashes['length'] == len(data)
    assert actual_hashes['checksums']['sha1'] == 'fdd1ce606a904b08c816ba84f3125f2af44d92b2'  # noqa
    assert (actual_hashes['checksums']['sha256'] ==
            '1d9224378d77925d612c9f926eb9fb92850e6551def8328011b6a972323298d5')


@pytest.mark.fs
def test_download_headers(tmp_path, requests_mock):
    """Check that we send proper headers when downloading files"""
    filename = 'requests-0.0.1.tar.gz'
    url = 'https://pypi.org/pypi/requests/%s' % filename
    data = 'this is something'
    requests_mock.get(url, text=data, headers={
        'content-length': str(len(data))
    })

    actual_filepath, actual_hashes = download(url, dest=str(tmp_path))

    assert len(requests_mock.request_history) == 1
    req = requests_mock.request_history[0]
    assert 'User-Agent' in req.headers
    user_agent = req.headers['User-Agent']
    assert 'Software Heritage Loader' in user_agent
    assert swh.loader.package.__version__ in user_agent


@pytest.mark.fs
def test_download_ok_with_hashes(tmp_path, requests_mock):
    """Download without issue should provide filename and hashes"""
    filename = 'requests-0.0.1.tar.gz'
    url = 'https://pypi.org/pypi/requests/%s' % filename
    data = 'this is something'
    requests_mock.get(url, text=data, headers={
        'content-length': str(len(data))
    })

    # good hashes for such file
    good = {
        'sha1': 'fdd1ce606a904b08c816ba84f3125f2af44d92b2',
        'sha256': '1d9224378d77925d612c9f926eb9fb92850e6551def8328011b6a972323298d5',  # noqa
    }

    actual_filepath, actual_hashes = download(url, dest=str(tmp_path),
                                              hashes=good)

    actual_filename = os.path.basename(actual_filepath)
    assert actual_filename == filename
    assert actual_hashes['length'] == len(data)
    assert actual_hashes['checksums']['sha1'] == good['sha1']
    assert actual_hashes['checksums']['sha256'] == good['sha256']


@pytest.mark.fs
def test_download_fail_hashes_mismatch(tmp_path, requests_mock):
    """Mismatch hash after download should raise

    """
    filename = 'requests-0.0.1.tar.gz'
    url = 'https://pypi.org/pypi/requests/%s' % filename
    data = 'this is something'
    requests_mock.get(url, text=data, headers={
        'content-length': str(len(data))
    })

    # good hashes for such file
    good = {
        'sha1': 'fdd1ce606a904b08c816ba84f3125f2af44d92b2',
        'sha256': '1d9224378d77925d612c9f926eb9fb92850e6551def8328011b6a972323298d5',  # noqa
    }

    for hash_algo in good.keys():
        wrong_hash = good[hash_algo].replace('1', '0')
        expected_hashes = good.copy()
        expected_hashes[hash_algo] = wrong_hash  # set the wrong hash

        expected_msg = ("Failure when fetching %s. "
                        "Checksum mismatched: %s != %s" % (
                            url, wrong_hash, good[hash_algo]
                        ))

        with pytest.raises(ValueError, match=expected_msg):
            download(url, dest=str(tmp_path), hashes=expected_hashes)


def test_api_info_failure(requests_mock):
    """Failure to fetch info/release information should raise"""
    url = 'https://pypi.org/pypi/requests/json'
    status_code = 400
    requests_mock.get(url, status_code=status_code)

    with pytest.raises(ValueError) as e0:
        api_info(url)

    assert e0.value.args[0] == "Fail to query '%s'. Reason: %s" % (
        url, status_code
    )


def test_api_info(requests_mock):
    """Fetching json info from pypi project should be ok"""
    url = 'https://pypi.org/pypi/requests/json'
    requests_mock.get(url, text='{"version": "0.0.1"}')
    actual_info = api_info(url)
    assert actual_info == {
        'version': '0.0.1',
    }


def test_release_name():
    for version, filename, expected_release in [
            ('0.0.1', None, 'releases/0.0.1'),
            ('0.0.2', 'something', 'releases/0.0.2/something')]:
        assert release_name(version, filename) == expected_release



# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import email.utils
import iso8601
import logging
import re
import subprocess

from dateutil.parser import parse as parse_date
from debian.changelog import Changelog
from debian.deb822 import Dsc
from os import path
from typing import (
    Any, Dict, Generator, List, Mapping, Optional, Sequence, Tuple
)

from swh.loader.package.loader import PackageLoader
from swh.loader.package.utils import download, release_name


logger = logging.getLogger(__name__)
UPLOADERS_SPLIT = re.compile(r'(?<=\>)\s*,\s*')


class DebianLoader(PackageLoader):
    """Load debian origins into swh archive.

    """
    visit_type = 'deb'

    def __init__(self, url: str, date: str, packages: Mapping[str, Any]):
        """Debian Loader implementation.

        Args:
            url: Origin url (e.g. deb://Debian/packages/cicero)
            date: Ignored
            packages: versioned packages and associated artifacts, example::

              {
                'stretch/contrib/0.7.2-3': {
                  'name': 'cicero',
                  'version': '0.7.2-3'
                  'files': {
                    'cicero_0.7.2-3.diff.gz': {
                       'md5sum': 'a93661b6a48db48d59ba7d26796fc9ce',
                       'name': 'cicero_0.7.2-3.diff.gz',
                       'sha256': 'f039c9642fe15c75bed5254315e2a29f...',
                       'size': 3964,
                       'uri': 'http://d.d.o/cicero_0.7.2-3.diff.gz',
                    },
                    'cicero_0.7.2-3.dsc': {
                      'md5sum': 'd5dac83eb9cfc9bb52a15eb618b4670a',
                      'name': 'cicero_0.7.2-3.dsc',
                      'sha256': '35b7f1048010c67adfd8d70e4961aefb...',
                      'size': 1864,
                      'uri': 'http://d.d.o/cicero_0.7.2-3.dsc',
                    },
                    'cicero_0.7.2.orig.tar.gz': {
                      'md5sum': '4353dede07c5728319ba7f5595a7230a',
                      'name': 'cicero_0.7.2.orig.tar.gz',
                      'sha256': '63f40f2436ea9f67b44e2d4bd669dbab...',
                      'size': 96527,
                      'uri': 'http://d.d.o/cicero_0.7.2.orig.tar.gz',
                    }
                  },
                },
                # ...
              }

        """
        super().__init__(url=url)
        self.packages = packages

    def get_versions(self) -> Sequence[str]:
        """Returns the keys of the packages input (e.g.
           stretch/contrib/0.7.2-3, etc...)

        """
        return list(self.packages.keys())

    def get_package_info(self, version: str) -> Generator[
            Tuple[str, Mapping[str, Any]], None, None]:
        meta = self.packages[version]
        p_info = meta.copy()
        p_info['raw'] = meta
        yield release_name(version), p_info

    def resolve_revision_from(
            self, known_package_artifacts: Dict, artifact_metadata: Dict) \
            -> Optional[bytes]:
        artifacts_to_fetch = artifact_metadata['files']
        logger.debug('k_p_artifacts: %s', known_package_artifacts)
        logger.debug('artifacts_to_fetch: %s', artifacts_to_fetch)
        for rev_id, known_artifacts in known_package_artifacts.items():
            logger.debug('Revision: %s', rev_id)
            logger.debug('Associated known_artifacts: %s', known_artifacts)
            known_artifacts = known_artifacts['extrinsic']['raw']['files']
            rev_found = True
            for a_name, k_artifact in known_artifacts.items():
                artifact_to_fetch = artifacts_to_fetch.get(a_name)
                logger.debug('artifact_to_fetch: %s', artifact_to_fetch)
                if artifact_to_fetch is None:
                    # as soon as we do not see an artifact, we consider we need
                    # to check the other revision
                    rev_found = False
                if k_artifact['sha256'] != artifact_to_fetch['sha256']:
                    # Hash is different, we consider we need to check the other
                    # revisions
                    rev_found = False
            if rev_found:
                logger.debug('Existing revision %s found for new artifacts.',
                             rev_id)
                return rev_id
        logger.debug('No existing revision found for the new artifacts.')
        return None

    def download_package(self, p_info: Mapping[str, Any],
                         tmpdir: str) -> List[Tuple[str, Mapping]]:
        """Contrary to other package loaders (1 package, 1 artifact),
        `a_metadata` represents the package's datafiles set to fetch:
        - <package-version>.orig.tar.gz
        - <package-version>.dsc
        - <package-version>.diff.gz

        This is delegated to the `download_package` function.

        """
        all_hashes = download_package(p_info, tmpdir)
        logger.debug('all_hashes: %s', all_hashes)
        res = []
        for hashes in all_hashes.values():
            res.append((tmpdir, hashes))
            logger.debug('res: %s', res)
        return res

    def uncompress(self, dl_artifacts: List[Tuple[str, Mapping[str, Any]]],
                   dest: str) -> str:
        logger.debug('dl_artifacts: %s', dl_artifacts)
        return extract_package(dl_artifacts, dest=dest)

    def build_revision(self, a_metadata: Mapping[str, Any],
                       uncompressed_path: str) -> Dict:
        dsc_url, dsc_name = dsc_information(a_metadata)
        if not dsc_name:
            raise ValueError(
                'dsc name for url %s should not be None' % dsc_url)
        dsc_path = path.join(path.dirname(uncompressed_path), dsc_name)
        i_metadata = get_package_metadata(
            a_metadata, dsc_path, uncompressed_path)

        logger.debug('i_metadata: %s', i_metadata)
        logger.debug('a_metadata: %s', a_metadata)

        msg = 'Synthetic revision for Debian source package %s version %s' % (
            a_metadata['name'], a_metadata['version'])

        date = iso8601.parse_date(i_metadata['changelog']['date'])
        author = prepare_person(i_metadata['changelog']['person'])

        # inspired from swh.loader.debian.converters.package_metadata_to_revision  # noqa
        return {
            'type': 'dsc',
            'message': msg.encode('utf-8'),
            'author': author,
            'date': date,
            'committer': author,
            'committer_date': date,
            'parents': [],
            'metadata': {
                'intrinsic': {
                    'tool': 'dsc',
                    'raw': i_metadata,
                },
                'extrinsic': {
                    'provider': dsc_url,
                    'when': self.visit_date.isoformat(),
                    'raw': a_metadata,
                },
            }
        }


def uid_to_person(uid: str) -> Mapping[str, str]:
    """Convert an uid to a person suitable for insertion.

    Args:
        uid: an uid of the form "Name <email@ddress>"

    Returns:
        a dictionary with the following keys:

        - name: the name associated to the uid
        - email: the mail associated to the uid
        - fullname: the actual uid input

    """
    logger.debug('uid: %s', uid)
    ret = {
        'name': '',
        'email': '',
        'fullname': uid,
    }

    name, mail = email.utils.parseaddr(uid)
    if name and email:
        ret['name'] = name
        ret['email'] = mail
    else:
        ret['name'] = uid
    return ret


def prepare_person(person: Mapping[str, str]) -> Mapping[str, bytes]:
    """Prepare person for swh serialization...

    Args:
        A person dict

    Returns:
        A person dict ready for storage

    """
    ret = {}
    for key, value in person.items():
        ret[key] = value.encode('utf-8')
    return ret


def download_package(
        package: Mapping[str, Any], tmpdir: Any) -> Mapping[str, Any]:
    """Fetch a source package in a temporary directory and check the checksums
    for all files.

    Args:
        package: Dict defining the set of files representing a debian package
        tmpdir: Where to download and extract the files to ingest

    Returns:
        Dict of swh hashes per filename key

    """
    all_hashes = {}
    for filename, fileinfo in package['files'].items():
        uri = fileinfo['uri']
        logger.debug('fileinfo: %s', fileinfo)
        extrinsic_hashes = {'sha256': fileinfo['sha256']}
        logger.debug('extrinsic_hashes(%s): %s', filename, extrinsic_hashes)
        filepath, hashes = download(uri, dest=tmpdir, filename=filename,
                                    hashes=extrinsic_hashes)
        all_hashes[filename] = hashes

    logger.debug('all_hashes: %s', all_hashes)
    return all_hashes


def dsc_information(package: Mapping[str, Any]) -> Tuple[
        Optional[str], Optional[str]]:
    """Retrieve dsc information from a package.

    Args:
        package: Package metadata information

    Returns:
        Tuple of dsc file's uri, dsc's full disk path

    """
    dsc_name = None
    dsc_url = None
    for filename, fileinfo in package['files'].items():
        if filename.endswith('.dsc'):
            if dsc_name:
                raise ValueError(
                    'Package %s_%s references several dsc files.' %
                    (package['name'], package['version'])
                )
            dsc_url = fileinfo['uri']
            dsc_name = filename

    return dsc_url, dsc_name


def extract_package(dl_artifacts: List[Tuple[str, Mapping]], dest: str) -> str:
    """Extract a Debian source package to a given directory.

    Note that after extraction the target directory will be the root of the
    extracted package, rather than containing it.

    Args:
        package: package information dictionary
        dest: directory where the package files are stored

    Returns:
        Package extraction directory

    """
    a_path = dl_artifacts[0][0]
    logger.debug('dl_artifacts: %s', dl_artifacts)
    for _, hashes in dl_artifacts:
        logger.debug('hashes: %s', hashes)
        filename = hashes['filename']
        if filename.endswith('.dsc'):
            dsc_name = filename
            break

    dsc_path = path.join(a_path, dsc_name)
    destdir = path.join(dest, 'extracted')
    logfile = path.join(dest, 'extract.log')
    logger.debug('extract Debian source package %s in %s' %
                 (dsc_path, destdir), extra={
                     'swh_type': 'deb_extract',
                     'swh_dsc': dsc_path,
                     'swh_destdir': destdir,
                 })

    cmd = ['dpkg-source',
           '--no-copy', '--no-check',
           '--ignore-bad-version',
           '-x', dsc_path,
           destdir]

    try:
        with open(logfile, 'w') as stdout:
            subprocess.check_call(cmd, stdout=stdout, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        logdata = open(logfile, 'r').read()
        raise ValueError('dpkg-source exited with code %s: %s' %
                         (e.returncode, logdata)) from None

    return destdir


def get_package_metadata(package: Mapping[str, Any], dsc_path: str,
                         extracted_path: str) -> Mapping[str, Any]:
    """Get the package metadata from the source package at dsc_path,
    extracted in extracted_path.

    Args:
        package: the package dict (with a dsc_path key)
        dsc_path: path to the package's dsc file
        extracted_path: the path where the package got extracted

    Returns:
        dict: a dictionary with the following keys:

        - history: list of (package_name, package_version) tuples parsed from
          the package changelog

    """
    with open(dsc_path, 'rb') as dsc:
        parsed_dsc = Dsc(dsc)

    # Parse the changelog to retrieve the rest of the package information
    changelog_path = path.join(extracted_path, 'debian/changelog')
    with open(changelog_path, 'rb') as changelog:
        try:
            parsed_changelog = Changelog(changelog)
        except UnicodeDecodeError:
            logger.warning('Unknown encoding for changelog %s,'
                           ' falling back to iso' %
                           changelog_path, extra={
                               'swh_type': 'deb_changelog_encoding',
                               'swh_name': package['name'],
                               'swh_version': str(package['version']),
                               'swh_changelog': changelog_path,
                           })

            # need to reset as Changelog scrolls to the end of the file
            changelog.seek(0)
            parsed_changelog = Changelog(changelog, encoding='iso-8859-15')

    package_info = {
        'name': package['name'],
        'version': str(package['version']),
        'changelog': {
            'person': uid_to_person(parsed_changelog.author),
            'date': parse_date(parsed_changelog.date).isoformat(),
            'history': [(block.package, str(block.version))
                        for block in parsed_changelog][1:],
        }
    }

    maintainers = [
        uid_to_person(parsed_dsc['Maintainer']),
    ]
    maintainers.extend(
        uid_to_person(person)
        for person in UPLOADERS_SPLIT.split(parsed_dsc.get('Uploaders', ''))
    )
    package_info['maintainers'] = maintainers

    return package_info



# Copyright (C) 2015-2016  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


class QueuePerNbElements():
    """Basic queue which holds the nb of elements it contains.

    """
    def __init__(self, max_nb_elements):
        self.reset()
        self.max_nb_elements = max_nb_elements

    def add(self, elements):
        if not isinstance(elements, list):
            elements = list(elements)
        self.elements.extend(elements)
        self.count += len(elements)
        return self.count >= self.max_nb_elements

    def pop(self):
        elements = self.elements
        self.reset()
        return elements

    def reset(self):
        self.elements = []
        self.count = 0


class QueuePerSizeAndNbUniqueElements():
    """Queue which permits to add unknown elements and holds the current
       size of the queue.

    """
    def __init__(self, max_nb_elements, max_size, key):
        self.reset()
        self.max_nb_elements = max_nb_elements
        self.max_size = max_size
        self.key = key
        self.keys = set()

    def _add_element(self, e):
        k = e[self.key]
        if k not in self.keys:
            self.keys.add(k)
            self.elements.append(e)
            self.size += e['length']
            self.count += 1

    def add(self, elements):
        for e in elements:
            self._add_element(e)
        return self.size >= self.max_size or \
            self.count >= self.max_nb_elements

    def pop(self):
        elements = self.elements
        self.reset()
        return elements

    def reset(self):
        self.elements = []
        self.keys = set()
        self.size = 0
        self.count = 0


class QueuePerNbUniqueElements():
    """Queue which permits to add unknown elements and knows the actual
       count of elements it held.

    """
    def __init__(self, max_nb_elements, key):
        self.reset()
        self.max_nb_elements = max_nb_elements
        self.key = key
        self.keys = set()

    def _add_element(self, e):
        k = e[self.key]
        if k not in self.keys:
            self.keys.add(k)
            self.elements.append(e)
            self.count += 1

    def add(self, elements):
        for e in elements:
            self._add_element(e)
        return self.count >= self.max_nb_elements

    def pop(self):
        elements = self.elements
        self.reset()
        return elements

    def reset(self):
        self.elements = []
        self.keys = set()
        self.count = 0



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import hashlib
import logging
import os
import psycopg2
import requests
import traceback
import uuid

from abc import ABCMeta, abstractmethod
from retrying import retry
from typing import Any, Dict, Optional, Tuple

from . import converters

from swh.core import config
from swh.storage import get_storage, HashCollision

from .queue import QueuePerSizeAndNbUniqueElements
from .queue import QueuePerNbUniqueElements


def send_in_packets(objects, sender, packet_size, packet_size_bytes=None):
    """Send `objects`, using the `sender`, in packets of `packet_size` objects
    (and of max `packet_size_bytes`).
    """
    formatted_objects = []
    count = 0
    if not packet_size_bytes:
        packet_size_bytes = 0
    for obj in objects:
        if not obj:
            continue
        formatted_objects.append(obj)
        if packet_size_bytes:
            count += obj['length']
        if len(formatted_objects) >= packet_size or count > packet_size_bytes:
            sender(formatted_objects)
            formatted_objects = []
            count = 0

    if formatted_objects:
        sender(formatted_objects)


def retry_loading(error):
    """Retry policy when the database raises an integrity error"""
    exception_classes = [
        # raised when two parallel insertions insert the same data.
        psycopg2.IntegrityError,
        HashCollision,
        # raised when uWSGI restarts and hungs up on the worker.
        requests.exceptions.ConnectionError,
    ]

    if not any(isinstance(error, exc) for exc in exception_classes):
        return False

    logger = logging.getLogger('swh.loader')

    error_name = error.__module__ + '.' + error.__class__.__name__
    logger.warning('Retry loading a batch', exc_info=False, extra={
        'swh_type': 'storage_retry',
        'swh_exception_type': error_name,
        'swh_exception': traceback.format_exception(
            error.__class__,
            error,
            error.__traceback__,
        ),
    })

    return True


class BufferedLoader(config.SWHConfig, metaclass=ABCMeta):
    """Mixin base class for loader.

    To use this class, you must:

    - inherit from this class
    - and implement the @abstractmethod methods:

      - :func:`prepare`: First step executed by the loader to prepare some
        state needed by the `func`:load method.

      - :func:`get_origin`: Retrieve the origin that is currently being loaded.

      - :func:`fetch_data`: Fetch the data is actually the method to implement
        to compute data to inject in swh (through the store_data method)

      - :func:`store_data`: Store data fetched.

      - :func:`visit_status`: Explicit status of the visit ('partial' or
        'full')

      - :func:`load_status`: Explicit status of the loading, for use by the
        scheduler (eventful/uneventful/temporary failure/permanent failure).

      - :func:`cleanup`: Last step executed by the loader.

    The entry point for the resulting loader is :func:`load`.

    You can take a look at some example classes:

    - :class:`BaseSvnLoader`
    - :class:`TarLoader`
    - :class:`DirLoader`
    - :class:`DebianLoader`

    """
    CONFIG_BASE_FILENAME = None  # type: Optional[str]

    DEFAULT_CONFIG = {
        'storage': ('dict', {
            'cls': 'remote',
            'args': {
                'url': 'http://localhost:5002/',
            }
        }),

        'send_contents': ('bool', True),
        'send_directories': ('bool', True),
        'send_revisions': ('bool', True),
        'send_releases': ('bool', True),
        'send_snapshot': ('bool', True),

        'save_data': ('bool', False),
        'save_data_path': ('str', ''),

        # Number of contents
        'content_packet_size': ('int', 10000),
        # packet of 100Mib contents
        'content_packet_size_bytes': ('int', 100 * 1024 * 1024),
        'directory_packet_size': ('int', 25000),
        'revision_packet_size': ('int', 100000),
        'release_packet_size': ('int', 100000),
        'occurrence_packet_size': ('int', 100000),
    }  # type: Dict[str, Tuple[str, Any]]

    ADDITIONAL_CONFIG = {}  # type: Dict[str, Tuple[str, Any]]

    def __init__(self, logging_class=None, config=None):
        if config:
            self.config = config
        else:
            self.config = self.parse_config_file(
                additional_configs=[self.ADDITIONAL_CONFIG])

        self.storage = get_storage(**self.config['storage'])

        if logging_class is None:
            logging_class = '%s.%s' % (self.__class__.__module__,
                                       self.__class__.__name__)
        self.log = logging.getLogger(logging_class)

        self.contents = QueuePerSizeAndNbUniqueElements(
            key='sha1',
            max_nb_elements=self.config['content_packet_size'],
            max_size=self.config['content_packet_size_bytes'])

        self.contents_seen = set()

        self.directories = QueuePerNbUniqueElements(
            key='id',
            max_nb_elements=self.config['directory_packet_size'])

        self.directories_seen = set()

        self.revisions = QueuePerNbUniqueElements(
            key='id',
            max_nb_elements=self.config['revision_packet_size'])

        self.revisions_seen = set()

        self.releases = QueuePerNbUniqueElements(
            key='id',
            max_nb_elements=self.config['release_packet_size'])

        self.releases_seen = set()

        self.snapshot = None

        _log = logging.getLogger('requests.packages.urllib3.connectionpool')
        _log.setLevel(logging.WARN)

        self.counters = {
            'contents': 0,
            'directories': 0,
            'revisions': 0,
            'releases': 0,
        }

        # Make sure the config is sane
        save_data = self.config.get('save_data')
        if save_data:
            path = self.config['save_data_path']
            os.stat(path)
            if not os.access(path, os.R_OK | os.W_OK):
                raise PermissionError("Permission denied: %r" % path)

    def save_data(self):
        """Save the data associated to the current load"""
        raise NotImplementedError

    def get_save_data_path(self):
        """The path to which we archive the loader's raw data"""
        if not hasattr(self, '__save_data_path'):
            year = str(self.visit_date.year)

            url = self.origin['url'].encode('utf-8')
            origin_url_hash = hashlib.sha1(url).hexdigest()

            path = '%s/sha1:%s/%s/%s' % (
                self.config['save_data_path'],
                origin_url_hash[0:2],
                origin_url_hash,
                year,
            )

            os.makedirs(path, exist_ok=True)
            self.__save_data_path = path

        return self.__save_data_path

    @retry(retry_on_exception=retry_loading, stop_max_attempt_number=3)
    def send_origin(self, origin):
        log_id = str(uuid.uuid4())
        self.log.debug('Creating origin for %s' % origin['url'],
                       extra={
                           'swh_type': 'storage_send_start',
                           'swh_content_type': 'origin',
                           'swh_num': 1,
                           'swh_id': log_id
                       })
        self.storage.origin_add_one(origin)
        self.log.debug('Done creating origin for %s' % origin['url'],
                       extra={
                           'swh_type': 'storage_send_end',
                           'swh_content_type': 'origin',
                           'swh_num': 1,
                           'swh_id': log_id
                       })

    @retry(retry_on_exception=retry_loading, stop_max_attempt_number=3)
    def send_origin_visit(self, visit_date, visit_type):
        log_id = str(uuid.uuid4())
        self.log.debug(
            'Creating origin_visit for origin %s at time %s' % (
                self.origin['url'], visit_date),
            extra={
                'swh_type': 'storage_send_start',
                'swh_content_type': 'origin_visit',
                'swh_num': 1,
                'swh_id': log_id
            })
        origin_visit = self.storage.origin_visit_add(
            self.origin['url'], visit_date, visit_type)
        self.log.debug(
            'Done Creating %s origin_visit for origin %s at time %s' % (
                visit_type, self.origin['url'], visit_date),
            extra={
                'swh_type': 'storage_send_end',
                'swh_content_type': 'origin_visit',
                'swh_num': 1,
                'swh_id': log_id
            })

        return origin_visit

    @retry(retry_on_exception=retry_loading, stop_max_attempt_number=3)
    def send_tool(self, tool):
        log_id = str(uuid.uuid4())
        self.log.debug(
            'Creating tool with name %s version %s configuration %s' % (
                 tool['name'], tool['version'], tool['configuration']),
            extra={
                'swh_type': 'storage_send_start',
                'swh_content_type': 'tool',
                'swh_num': 1,
                'swh_id': log_id
            })

        tools = self.storage.tool_add([tool])
        tool_id = tools[0]['id']

        self.log.debug(
            'Done creating tool with name %s version %s and configuration %s' % (  # noqa
                 tool['name'], tool['version'], tool['configuration']),
            extra={
                'swh_type': 'storage_send_end',
                'swh_content_type': 'tool',
                'swh_num': 1,
                'swh_id': log_id
            })
        return tool_id

    @retry(retry_on_exception=retry_loading, stop_max_attempt_number=3)
    def send_provider(self, provider):
        log_id = str(uuid.uuid4())
        self.log.debug(
            'Creating metadata_provider with name %s type %s url %s' % (
                provider['provider_name'], provider['provider_type'],
                provider['provider_url']),
            extra={
                'swh_type': 'storage_send_start',
                'swh_content_type': 'metadata_provider',
                'swh_num': 1,
                'swh_id': log_id
            })
        # FIXME: align metadata_provider_add with indexer_configuration_add
        _provider = self.storage.metadata_provider_get_by(provider)
        if _provider and 'id' in _provider:
            provider_id = _provider['id']
        else:
            provider_id = self.storage.metadata_provider_add(
                provider['provider_name'],
                provider['provider_type'],
                provider['provider_url'],
                provider['metadata'])

        self.log.debug(
            'Done creating metadata_provider with name %s type %s url %s' % (
                provider['provider_name'], provider['provider_type'],
                provider['provider_url']),
            extra={
                'swh_type': 'storage_send_end',
                'swh_content_type': 'metadata_provider',
                'swh_num': 1,
                'swh_id': log_id
            })
        return provider_id

    @retry(retry_on_exception=retry_loading, stop_max_attempt_number=3)
    def send_origin_metadata(self, visit_date, provider_id,
                             tool_id, metadata):
        log_id = str(uuid.uuid4())
        self.log.debug(
            'Creating origin_metadata for origin %s at time %s with provider_id %s and tool_id %s' % (  # noqa
                self.origin['url'], visit_date, provider_id, tool_id),
            extra={
                'swh_type': 'storage_send_start',
                'swh_content_type': 'origin_metadata',
                'swh_num': 1,
                'swh_id': log_id
            })

        self.storage.origin_metadata_add(
            self.origin['url'], visit_date, provider_id, tool_id, metadata)
        self.log.debug(
            'Done Creating origin_metadata for origin %s at time %s with provider %s and tool %s' % (  # noqa
                self.origin['url'], visit_date, provider_id, tool_id),
            extra={
                'swh_type': 'storage_send_end',
                'swh_content_type': 'origin_metadata',
                'swh_num': 1,
                'swh_id': log_id
            })

    @retry(retry_on_exception=retry_loading, stop_max_attempt_number=3)
    def update_origin_visit(self, status):
        log_id = str(uuid.uuid4())
        self.log.debug(
            'Updating origin_visit for origin %s with status %s' % (
                self.origin['url'], status),
            extra={
                'swh_type': 'storage_send_start',
                'swh_content_type': 'origin_visit',
                'swh_num': 1,
                'swh_id': log_id
            })
        self.storage.origin_visit_update(
            self.origin['url'], self.visit, status)
        self.log.debug(
            'Done updating origin_visit for origin %s with status %s' % (
                self.origin['url'], status),
            extra={
                'swh_type': 'storage_send_end',
                'swh_content_type': 'origin_visit',
                'swh_num': 1,
                'swh_id': log_id
            })

    @retry(retry_on_exception=retry_loading, stop_max_attempt_number=3)
    def send_contents(self, content_list):
        """Actually send properly formatted contents to the database.

        """
        num_contents = len(content_list)
        if num_contents > 0:
            log_id = str(uuid.uuid4())
            self.log.debug("Sending %d contents" % num_contents,
                           extra={
                               'swh_type': 'storage_send_start',
                               'swh_content_type': 'content',
                               'swh_num': num_contents,
                               'swh_id': log_id,
                           })
            result = self.storage.content_add(content_list)
            self.counters['contents'] += result.get('content:add', 0)
            self.log.debug("Done sending %d contents" % num_contents,
                           extra={
                               'swh_type': 'storage_send_end',
                               'swh_content_type': 'content',
                               'swh_num': num_contents,
                               'swh_id': log_id,
                           })

    @retry(retry_on_exception=retry_loading, stop_max_attempt_number=3)
    def send_directories(self, directory_list):
        """Actually send properly formatted directories to the database.

        """
        num_directories = len(directory_list)
        if num_directories > 0:
            log_id = str(uuid.uuid4())
            self.log.debug("Sending %d directories" % num_directories,
                           extra={
                               'swh_type': 'storage_send_start',
                               'swh_content_type': 'directory',
                               'swh_num': num_directories,
                               'swh_id': log_id,
                           })
            result = self.storage.directory_add(directory_list)
            self.counters['directories'] += result.get('directory:add', 0)
            self.log.debug("Done sending %d directories" % num_directories,
                           extra={
                               'swh_type': 'storage_send_end',
                               'swh_content_type': 'directory',
                               'swh_num': num_directories,
                               'swh_id': log_id,
                           })

    @retry(retry_on_exception=retry_loading, stop_max_attempt_number=3)
    def send_revisions(self, revision_list):
        """Actually send properly formatted revisions to the database.

        """
        num_revisions = len(revision_list)
        if num_revisions > 0:
            log_id = str(uuid.uuid4())
            self.log.debug("Sending %d revisions" % num_revisions,
                           extra={
                               'swh_type': 'storage_send_start',
                               'swh_content_type': 'revision',
                               'swh_num': num_revisions,
                               'swh_id': log_id,
                           })
            result = self.storage.revision_add(revision_list)
            self.counters['revisions'] += result.get('revision:add', 0)
            self.log.debug("Done sending %d revisions" % num_revisions,
                           extra={
                               'swh_type': 'storage_send_end',
                               'swh_content_type': 'revision',
                               'swh_num': num_revisions,
                               'swh_id': log_id,
                           })

    @retry(retry_on_exception=retry_loading, stop_max_attempt_number=3)
    def send_releases(self, release_list):
        """Actually send properly formatted releases to the database.

        """
        num_releases = len(release_list)
        if num_releases > 0:
            log_id = str(uuid.uuid4())
            self.log.debug("Sending %d releases" % num_releases,
                           extra={
                               'swh_type': 'storage_send_start',
                               'swh_content_type': 'release',
                               'swh_num': num_releases,
                               'swh_id': log_id,
                           })
            result = self.storage.release_add(release_list)
            self.counters['releases'] += result.get('release:add', 0)
            self.log.debug("Done sending %d releases" % num_releases,
                           extra={
                               'swh_type': 'storage_send_end',
                               'swh_content_type': 'release',
                               'swh_num': num_releases,
                               'swh_id': log_id,
                           })

    @retry(retry_on_exception=retry_loading, stop_max_attempt_number=3)
    def send_snapshot(self, snapshot):
        self.storage.snapshot_add([snapshot])
        self.storage.origin_visit_update(
            self.origin['url'], self.visit, snapshot=snapshot['id'])

    @retry(retry_on_exception=retry_loading, stop_max_attempt_number=3)
    def filter_missing_contents(self, contents):
        """Return only the contents missing from swh"""
        max_content_size = self.config['content_size_limit']
        contents_per_key = {}
        content_key = 'blake2s256'

        for content in contents:
            if content[content_key] in self.contents_seen:
                continue
            key = content[content_key]
            contents_per_key[key] = content
            self.contents_seen.add(key)

        for key in self.storage.content_missing(
                list(contents_per_key.values()),
                key_hash=content_key
        ):
            yield converters.content_for_storage(
                contents_per_key[key],
                max_content_size=max_content_size,
                origin_url=self.origin['url'],
            )

    def bulk_send_contents(self, contents):
        """Format contents as swh contents and send them to the database.

        """
        threshold_reached = self.contents.add(
            self.filter_missing_contents(contents))
        if threshold_reached:
            self.send_batch_contents(self.contents.pop())

    def filter_missing_directories(self, directories):
        """Return only directories missing from swh"""

        directories_per_id = {}
        search_dirs = []

        for directory in directories:
            dir_id = directory['id']
            if dir_id in self.directories_seen:
                continue

            search_dirs.append(dir_id)
            directories_per_id[dir_id] = directory
            self.directories_seen.add(dir_id)

        for dir_id in self.storage.directory_missing(search_dirs):
            yield directories_per_id[dir_id]

    def bulk_send_directories(self, directories):
        """Send missing directories to the database"""
        threshold_reached = self.directories.add(
            self.filter_missing_directories(directories))
        if threshold_reached:
            self.send_batch_contents(self.contents.pop())
            self.send_batch_directories(self.directories.pop())

    def filter_missing_revisions(self, revisions):
        """Return only revisions missing from swh"""
        revisions_per_id = {}
        search_revs = []

        for revision in revisions:
            rev_id = revision['id']
            if rev_id in self.revisions_seen:
                continue

            search_revs.append(rev_id)
            revisions_per_id[rev_id] = revision
            self.revisions_seen.add(rev_id)

        for rev_id in self.storage.revision_missing(search_revs):
            yield revisions_per_id[rev_id]

    def bulk_send_revisions(self, revisions):
        """Send missing revisions to the database"""
        threshold_reached = self.revisions.add(
            self.filter_missing_revisions(revisions))
        if threshold_reached:
            self.send_batch_contents(self.contents.pop())
            self.send_batch_directories(self.directories.pop())
            self.send_batch_revisions(self.revisions.pop())

    def filter_missing_releases(self, releases):
        """Return only releases missing from swh"""
        releases_per_id = {}
        search_rels = []

        for release in releases:
            rel_id = release['id']
            if rel_id in self.releases_seen:
                continue

            search_rels.append(rel_id)
            releases_per_id[rel_id] = release
            self.releases_seen.add(rel_id)

        for rel_id in self.storage.release_missing(search_rels):
            yield releases_per_id[rel_id]

    def bulk_send_releases(self, releases):
        """Send missing releases to the database"""
        threshold_reached = self.releases.add(
            self.filter_missing_releases(releases))
        if threshold_reached:
            self.send_batch_contents(self.contents.pop())
            self.send_batch_directories(self.directories.pop())
            self.send_batch_revisions(self.revisions.pop())
            self.send_batch_releases(self.releases.pop())

    def bulk_send_snapshot(self, snapshot):
        """Send missing releases to the database"""
        self.send_batch_contents(self.contents.pop())
        self.send_batch_directories(self.directories.pop())
        self.send_batch_revisions(self.revisions.pop())
        self.send_batch_releases(self.releases.pop())
        self.send_snapshot(snapshot)

    def maybe_load_contents(self, contents):
        """Load contents in swh-storage if need be.

        """
        if self.config['send_contents']:
            self.bulk_send_contents(contents)

    def maybe_load_directories(self, directories):
        """Load directories in swh-storage if need be.

        """
        if self.config['send_directories']:
            self.bulk_send_directories(directories)

    def maybe_load_revisions(self, revisions):
        """Load revisions in swh-storage if need be.

        """
        if self.config['send_revisions']:
            self.bulk_send_revisions(revisions)

    def maybe_load_releases(self, releases):
        """Load releases in swh-storage if need be.

        """
        if self.config['send_releases']:
            self.bulk_send_releases(releases)

    def maybe_load_snapshot(self, snapshot):
        """Load the snapshot in swh-storage if need be."""
        if self.config['send_snapshot']:
            self.bulk_send_snapshot(snapshot)

    def send_batch_contents(self, contents):
        """Send contents batches to the storage"""
        packet_size = self.config['content_packet_size']
        packet_size_bytes = self.config['content_packet_size_bytes']
        send_in_packets(contents, self.send_contents, packet_size,
                        packet_size_bytes=packet_size_bytes)

    def send_batch_directories(self, directories):
        """Send directories batches to the storage"""
        packet_size = self.config['directory_packet_size']
        send_in_packets(directories, self.send_directories, packet_size)

    def send_batch_revisions(self, revisions):
        """Send revisions batches to the storage"""
        packet_size = self.config['revision_packet_size']
        send_in_packets(revisions, self.send_revisions, packet_size)

    def send_batch_releases(self, releases):
        """Send releases batches to the storage
        """
        packet_size = self.config['release_packet_size']
        send_in_packets(releases, self.send_releases, packet_size)

    def flush(self):
        """Flush any potential dangling data not sent to swh-storage.

        Bypass the maybe_load_* methods which awaits threshold reached
        signal. We actually want to store those as we are done
        loading.

        """
        contents = self.contents.pop()
        directories = self.directories.pop()
        revisions = self.revisions.pop()
        releases = self.releases.pop()

        # and send those to storage if asked
        if self.config['send_contents']:
            self.send_batch_contents(contents)
        if self.config['send_contents']:
            self.send_batch_directories(directories)
        if self.config['send_revisions']:
            self.send_batch_revisions(revisions)
        if self.config['send_releases']:
            self.send_batch_releases(releases)
        if self.config['send_snapshot'] and self.snapshot:
            self.send_snapshot(self.snapshot)
        if hasattr(self.storage, 'flush'):
            self.storage.flush()

    def prepare_metadata(self):
        """First step for origin_metadata insertion, resolving the
        provider_id and the tool_id by fetching data from the storage
        or creating tool and provider on the fly if the data isn't available

        """
        origin_metadata = self.origin_metadata

        tool = origin_metadata['tool']
        try:
            tool_id = self.send_tool(tool)
            self.origin_metadata['tool']['tool_id'] = tool_id
        except Exception:
            self.log.exception('Problem when storing new tool')
            raise

        provider = origin_metadata['provider']
        try:
            provider_id = self.send_provider(provider)
            self.origin_metadata['provider']['provider_id'] = provider_id
        except Exception:
            self.log.exception('Problem when storing new provider')
            raise

    @abstractmethod
    def cleanup(self):
        """Last step executed by the loader.

        """
        pass

    @abstractmethod
    def prepare_origin_visit(self, *args, **kwargs):
        """First step executed by the loader to prepare origin and visit
           references. Set/update self.origin, and
           optionally self.origin_url, self.visit_date.

        """
        pass

    def _store_origin_visit(self):
        """Store origin and visit references. Sets the self.origin_visit and
           self.visit references.

        """
        origin = self.origin.copy()
        self.send_origin(origin)

        if not self.visit_date:  # now as default visit_date if not provided
            self.visit_date = datetime.datetime.now(tz=datetime.timezone.utc)
        self.origin_visit = self.send_origin_visit(
            self.visit_date, self.visit_type)
        self.visit = self.origin_visit['visit']

    @abstractmethod
    def prepare(self, *args, **kwargs):
        """Second step executed by the loader to prepare some state needed by
           the loader.

        """
        pass

    def get_origin(self):
        """Get the origin that is currently being loaded.
        self.origin should be set in :func:`prepare_origin`

        Returns:
          dict: an origin ready to be sent to storage by
          :func:`origin_add_one`.
        """
        return self.origin

    @abstractmethod
    def fetch_data(self):
        """Fetch the data from the source the loader is currently loading
           (ex: git/hg/svn/... repository).

        Returns:
            a value that is interpreted as a boolean. If True, fetch_data needs
            to be called again to complete loading.

        """
        pass

    @abstractmethod
    def store_data(self):
        """Store fetched data in the database.

        Should call the :func:`maybe_load_xyz` methods, which handle the
        bundles sent to storage, rather than send directly.
        """
        pass

    def store_metadata(self):
        """Store fetched metadata in the database.

        For more information, see implementation in :class:`DepositLoader`.
        """
        pass

    def load_status(self):
        """Detailed loading status.

        Defaults to logging an eventful load.

        Returns: a dictionary that is eventually passed back as the task's
          result to the scheduler, allowing tuning of the task recurrence
          mechanism.
        """
        return {
            'status': 'eventful',
        }

    def post_load(self, success=True):
        """Permit the loader to do some additional actions according to status
        after the loading is done. The flag success indicates the
        loading's status.

        Defaults to doing nothing.

        This is up to the implementer of this method to make sure this
        does not break.

        Args:
            success (bool): the success status of the loading

        """
        pass

    def visit_status(self):
        """Detailed visit status.

        Defaults to logging a full visit.
        """
        return 'full'

    def pre_cleanup(self):
        """As a first step, will try and check for dangling data to cleanup.
        This should do its best to avoid raising issues.

        """
        pass

    def load(self, *args, **kwargs):
        r"""Loading logic for the loader to follow:

        - 1. Call :meth:`prepare_origin_visit` to prepare the
             origin and visit we will associate loading data to
        - 2. Store the actual ``origin_visit`` to storage
        - 3. Call :meth:`prepare` to prepare any eventual state
        - 4. Call :meth:`get_origin` to get the origin we work with and store

        - while True:

          - 5. Call :meth:`fetch_data` to fetch the data to store
          - 6. Call :meth:`store_data` to store the data

        - 7. Call :meth:`cleanup` to clean up any eventual state put in place
             in :meth:`prepare` method.

        """
        try:
            self.pre_cleanup()
        except Exception:
            msg = 'Cleaning up dangling data failed! Continue loading.'
            self.log.warning(msg)

        self.prepare_origin_visit(*args, **kwargs)
        self._store_origin_visit()

        try:
            self.prepare(*args, **kwargs)

            while True:
                more_data_to_fetch = self.fetch_data()
                self.store_data()
                if not more_data_to_fetch:
                    break

            self.store_metadata()
            self.update_origin_visit(status=self.visit_status())
            self.post_load()
        except Exception:
            self.log.exception('Loading failure, updating to `partial` status',
                               extra={
                                   'swh_task_args': args,
                                   'swh_task_kwargs': kwargs,
                               })
            self.update_origin_visit(status='partial')
            self.post_load(success=False)
            return {'status': 'failed'}
        finally:
            self.flush()
            self.cleanup()

        return self.load_status()


class UnbufferedLoader(BufferedLoader):
    """This base class is a pattern for unbuffered loaders.

    UnbufferedLoader loaders are able to load all the data in one go. For
    example, the loader defined in swh-loader-git
    :class:`BulkUpdater`.

    For other loaders (stateful one, (e.g :class:`SWHSvnLoader`),
    inherit directly from :class:`BufferedLoader`.

    """
    ADDITIONAL_CONFIG = {}  # type: Dict[str, Tuple[str, Any]]

    def __init__(self, logging_class=None, config=None):
        super().__init__(logging_class=logging_class, config=config)
        self.visit_date = None  # possibly overridden in self.prepare method

    def cleanup(self):
        """Clean up an eventual state installed for computations."""
        pass

    def has_contents(self):
        """Checks whether we need to load contents"""
        return True

    def get_contents(self):
        """Get the contents that need to be loaded"""
        raise NotImplementedError

    def has_directories(self):
        """Checks whether we need to load directories"""
        return True

    def get_directories(self):
        """Get the directories that need to be loaded"""
        raise NotImplementedError

    def has_revisions(self):
        """Checks whether we need to load revisions"""
        return True

    def get_revisions(self):
        """Get the revisions that need to be loaded"""
        raise NotImplementedError

    def has_releases(self):
        """Checks whether we need to load releases"""
        return True

    def get_releases(self):
        """Get the releases that need to be loaded"""
        raise NotImplementedError

    def get_snapshot(self):
        """Get the snapshot that needs to be loaded"""
        raise NotImplementedError

    def eventful(self):
        """Whether the load was eventful"""
        raise NotImplementedError

    def save_data(self):
        """Save the data associated to the current load"""
        raise NotImplementedError

    def flush(self):
        """Flush the storage if needed.
        """
        if hasattr(self.storage, 'flush'):
            self.storage.flush()

    def store_data(self):
        if self.config['save_data']:
            self.save_data()

        if self.config['send_contents'] and self.has_contents():
            self.send_batch_contents(self.get_contents())
        if self.config['send_directories'] and self.has_directories():
            self.send_batch_directories(self.get_directories())
        if self.config['send_revisions'] and self.has_revisions():
            self.send_batch_revisions(self.get_revisions())
        if self.config['send_releases'] and self.has_releases():
            self.send_batch_releases(self.get_releases())
        if self.config['send_snapshot']:
            self.send_snapshot(self.get_snapshot())



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import json
import os

from swh.model.hashutil import hash_to_bytes

from swh.loader.package.npm import (
    parse_npm_package_author, extract_npm_package_author
)
from swh.loader.package.tests.common import (
    check_snapshot, check_metadata_paths, get_stats
)

from swh.loader.package.npm import NpmLoader


def _parse_author_string_test(author_str, expected_result):
    assert parse_npm_package_author(author_str) == expected_result
    assert parse_npm_package_author(' %s' % author_str) == expected_result
    assert parse_npm_package_author('%s ' % author_str) == expected_result


def test_parse_npm_package_author():
    _parse_author_string_test(
        'John Doe',
        {
            'name': 'John Doe'
        }
    )

    _parse_author_string_test(
        '<john.doe@foo.bar>',
        {
            'email': 'john.doe@foo.bar'
        }
    )

    _parse_author_string_test(
        '(https://john.doe)',
        {
            'url': 'https://john.doe'
        }
    )

    _parse_author_string_test(
        'John Doe <john.doe@foo.bar>',
        {
            'name': 'John Doe',
            'email': 'john.doe@foo.bar'
        }
    )

    _parse_author_string_test(
        'John Doe<john.doe@foo.bar>',
        {
            'name': 'John Doe',
            'email': 'john.doe@foo.bar'
        }
    )

    _parse_author_string_test(
        'John Doe (https://john.doe)',
        {
            'name': 'John Doe',
            'url': 'https://john.doe'
        }
    )

    _parse_author_string_test(
        'John Doe(https://john.doe)',
        {
            'name': 'John Doe',
            'url': 'https://john.doe'
        }
    )

    _parse_author_string_test(
        '<john.doe@foo.bar> (https://john.doe)',
        {
            'email': 'john.doe@foo.bar',
            'url': 'https://john.doe'
        }
    )

    _parse_author_string_test(
        '(https://john.doe) <john.doe@foo.bar>',
        {
            'email': 'john.doe@foo.bar',
            'url': 'https://john.doe'
        }
    )

    _parse_author_string_test(
        'John Doe <john.doe@foo.bar> (https://john.doe)',
        {
            'name': 'John Doe',
            'email': 'john.doe@foo.bar',
            'url': 'https://john.doe'
        }
    )

    _parse_author_string_test(
        'John Doe (https://john.doe) <john.doe@foo.bar>',
        {
            'name': 'John Doe',
            'email': 'john.doe@foo.bar',
            'url': 'https://john.doe'
        }
    )

    _parse_author_string_test(
        'John Doe<john.doe@foo.bar> (https://john.doe)',
        {
            'name': 'John Doe',
            'email': 'john.doe@foo.bar',
            'url': 'https://john.doe'
        }
    )

    _parse_author_string_test(
        'John Doe<john.doe@foo.bar>(https://john.doe)',
        {
            'name': 'John Doe',
            'email': 'john.doe@foo.bar',
            'url': 'https://john.doe'
        }
    )

    _parse_author_string_test('', {})
    _parse_author_string_test('<>', {})
    _parse_author_string_test(' <>', {})
    _parse_author_string_test('<>()', {})
    _parse_author_string_test('<> ()', {})
    _parse_author_string_test('()', {})
    _parse_author_string_test(' ()', {})

    _parse_author_string_test(
        'John Doe <> ()',
        {
            'name': 'John Doe'
        }
    )

    _parse_author_string_test(
        'John Doe <>',
        {
            'name': 'John Doe'
        }
    )

    _parse_author_string_test(
        'John Doe ()',
        {
            'name': 'John Doe'
        }
    )


def test_extract_npm_package_author(datadir):
    package_metadata_filepath = os.path.join(
        datadir, 'https_replicate.npmjs.com', 'org_visit1')

    with open(package_metadata_filepath) as json_file:
        package_metadata = json.load(json_file)

    extract_npm_package_author(package_metadata['versions']['0.0.2']) == \
        {
            'fullname': b'mooz <stillpedant@gmail.com>',
            'name': b'mooz',
            'email': b'stillpedant@gmail.com'
        }

    assert (
        extract_npm_package_author(package_metadata['versions']['0.0.3']) ==
        {
            'fullname': b'Masafumi Oyamada <stillpedant@gmail.com>',
            'name': b'Masafumi Oyamada',
            'email': b'stillpedant@gmail.com'
        }
    )

    package_json = json.loads('''
    {
        "name": "highlightjs-line-numbers.js",
        "version": "2.7.0",
        "description": "Highlight.js line numbers plugin.",
        "main": "src/highlightjs-line-numbers.js",
        "dependencies": {},
        "devDependencies": {
            "gulp": "^4.0.0",
            "gulp-rename": "^1.4.0",
            "gulp-replace": "^0.6.1",
            "gulp-uglify": "^1.2.0"
        },
        "repository": {
            "type": "git",
            "url": "https://github.com/wcoder/highlightjs-line-numbers.js.git"
        },
        "author": "Yauheni Pakala <evgeniy.pakalo@gmail.com>",
        "license": "MIT",
        "bugs": {
            "url": "https://github.com/wcoder/highlightjs-line-numbers.js/issues"
        },
        "homepage": "http://wcoder.github.io/highlightjs-line-numbers.js/"
    }''') # noqa

    assert extract_npm_package_author(package_json) == \
        {
            'fullname': b'Yauheni Pakala <evgeniy.pakalo@gmail.com>',
            'name': b'Yauheni Pakala',
            'email': b'evgeniy.pakalo@gmail.com'
        }

    package_json = json.loads('''
    {
        "name": "3-way-diff",
        "version": "0.0.1",
        "description": "3-way diffing of JavaScript objects",
        "main": "index.js",
        "authors": [
            {
                "name": "Shawn Walsh",
                "url": "https://github.com/shawnpwalsh"
            },
            {
                "name": "Markham F Rollins IV",
                "url": "https://github.com/mrollinsiv"
            }
        ],
        "keywords": [
            "3-way diff",
            "3 way diff",
            "three-way diff",
            "three way diff"
        ],
        "devDependencies": {
            "babel-core": "^6.20.0",
            "babel-preset-es2015": "^6.18.0",
            "mocha": "^3.0.2"
        },
        "dependencies": {
            "lodash": "^4.15.0"
        }
    }''')

    assert extract_npm_package_author(package_json) == \
        {
            'fullname': b'Shawn Walsh',
            'name': b'Shawn Walsh',
            'email': None
        }

    package_json = json.loads('''
    {
        "name": "yfe-ynpm",
        "version": "1.0.0",
        "homepage": "http://gitlab.ywwl.com/yfe/yfe-ynpm",
        "repository": {
            "type": "git",
            "url": "git@gitlab.ywwl.com:yfe/yfe-ynpm.git"
        },
        "author": [
            "fengmk2 <fengmk2@gmail.com> (https://fengmk2.com)",
            "xufuzi <xufuzi@ywwl.com> (https://7993.org)"
        ],
        "license": "MIT"
    }''')

    assert extract_npm_package_author(package_json) == \
        {
            'fullname': b'fengmk2 <fengmk2@gmail.com>',
            'name': b'fengmk2',
            'email': b'fengmk2@gmail.com'
        }

    package_json = json.loads('''
    {
        "name": "umi-plugin-whale",
        "version": "0.0.8",
        "description": "Internal contract component",
        "authors": {
            "name": "xiaohuoni",
            "email": "448627663@qq.com"
        },
        "repository": "alitajs/whale",
        "devDependencies": {
            "np": "^3.0.4",
            "umi-tools": "*"
        },
        "license": "MIT"
    }''')

    assert extract_npm_package_author(package_json) == \
        {
            'fullname': b'xiaohuoni <448627663@qq.com>',
            'name': b'xiaohuoni',
            'email': b'448627663@qq.com'
        }


def normalize_hashes(hashes):
    if isinstance(hashes, str):
        return hash_to_bytes(hashes)
    if isinstance(hashes, list):
        return [hash_to_bytes(x) for x in hashes]
    return {hash_to_bytes(k): hash_to_bytes(v) for k, v in hashes.items()}


_expected_new_contents_first_visit = normalize_hashes([
    '4ce3058e16ab3d7e077f65aabf855c34895bf17c',
    '858c3ceee84c8311adc808f8cdb30d233ddc9d18',
    '0fa33b4f5a4e0496da6843a38ff1af8b61541996',
    '85a410f8ef8eb8920f2c384a9555566ad4a2e21b',
    '9163ac8025923d5a45aaac482262893955c9b37b',
    '692cf623b8dd2c5df2c2998fd95ae4ec99882fb4',
    '18c03aac6d3e910efb20039c15d70ab5e0297101',
    '41265c42446aac17ca769e67d1704f99e5a1394d',
    '783ff33f5882813dca9239452c4a7cadd4dba778',
    'b029cfb85107aee4590c2434a3329bfcf36f8fa1',
    '112d1900b4c2e3e9351050d1b542c9744f9793f3',
    '5439bbc4bd9a996f1a38244e6892b71850bc98fd',
    'd83097a2f994b503185adf4e719d154123150159',
    'd0939b4898e83090ee55fd9d8a60e312cfadfbaf',
    'b3523a26f7147e4af40d9d462adaae6d49eda13e',
    'cd065fb435d6fb204a8871bcd623d0d0e673088c',
    '2854a40855ad839a54f4b08f5cff0cf52fca4399',
    'b8a53bbaac34ebb8c6169d11a4b9f13b05c583fe',
    '0f73d56e1cf480bded8a1ecf20ec6fc53c574713',
    '0d9882b2dfafdce31f4e77fe307d41a44a74cefe',
    '585fc5caab9ead178a327d3660d35851db713df1',
    'e8cd41a48d79101977e3036a87aeb1aac730686f',
    '5414efaef33cceb9f3c9eb5c4cc1682cd62d14f7',
    '9c3cc2763bf9e9e37067d3607302c4776502df98',
    '3649a68410e354c83cd4a38b66bd314de4c8f5c9',
    'e96ed0c091de1ebdf587104eaf63400d1974a1fe',
    '078ca03d2f99e4e6eab16f7b75fbb7afb699c86c',
    '38de737da99514de6559ff163c988198bc91367a',
])

_expected_new_directories_first_visit = normalize_hashes([
    '3370d20d6f96dc1c9e50f083e2134881db110f4f',
    '42753c0c2ab00c4501b552ac4671c68f3cf5aece',
    'd7895533ef5edbcffdea3f057d9fef3a1ef845ce',
    '80579be563e2ef3e385226fe7a3f079b377f142c',
    '3b0ddc6a9e58b4b53c222da4e27b280b6cda591c',
    'bcad03ce58ac136f26f000990fc9064e559fe1c0',
    '5fc7e82a1bc72e074665c6078c6d3fad2f13d7ca',
    'e3cd26beba9b1e02f6762ef54bd9ac80cc5f25fd',
    '584b5b4b6cf7f038095e820b99386a9c232de931',
    '184c8d6d0d242f2b1792ef9d3bf396a5434b7f7a',
    'bb5f4ee143c970367eb409f2e4c1104898048b9d',
    '1b95491047add1103db0dfdfa84a9735dcb11e88',
    'a00c6de13471a2d66e64aca140ddb21ef5521e62',
    '5ce6c1cd5cda2d546db513aaad8c72a44c7771e2',
    'c337091e349b6ac10d38a49cdf8c2401ef9bb0f2',
    '202fafcd7c0f8230e89d5496ad7f44ab12b807bf',
    '775cc516543be86c15c1dc172f49c0d4e6e78235',
    'ff3d1ead85a14f891e8b3fa3a89de39db1b8de2e',
])

_expected_new_revisions_first_visit = normalize_hashes({
    'd8a1c7474d2956ac598a19f0f27d52f7015f117e':
    '42753c0c2ab00c4501b552ac4671c68f3cf5aece',
    '5f9eb78af37ffd12949f235e86fac04898f9f72a':
    '3370d20d6f96dc1c9e50f083e2134881db110f4f',
    'ba019b192bdb94bd0b5bd68b3a5f92b5acc2239a':
    'd7895533ef5edbcffdea3f057d9fef3a1ef845ce'}
)


def package_url(package):
    return 'https://www.npmjs.com/package/%s' % package


def package_metadata_url(package):
    return 'https://replicate.npmjs.com/%s/' % package


def test_revision_metadata_structure(swh_config, requests_mock_datadir):
    package = 'org'
    loader = NpmLoader(package,
                       package_url(package),
                       package_metadata_url(package))

    actual_load_status = loader.load()
    assert actual_load_status['status'] == 'eventful'
    assert actual_load_status['snapshot_id'] is not None

    expected_revision_id = hash_to_bytes(
        'd8a1c7474d2956ac598a19f0f27d52f7015f117e')
    revision = list(loader.storage.revision_get([expected_revision_id]))[0]

    assert revision is not None

    check_metadata_paths(revision['metadata'], paths=[
        ('intrinsic.tool', str),
        ('intrinsic.raw', dict),
        ('extrinsic.provider', str),
        ('extrinsic.when', str),
        ('extrinsic.raw', dict),
        ('original_artifact', list),
    ])

    for original_artifact in revision['metadata']['original_artifact']:
        check_metadata_paths(original_artifact, paths=[
            ('filename', str),
            ('length', int),
            ('checksums', dict),
        ])


def test_npm_loader_first_visit(swh_config, requests_mock_datadir):

    package = 'org'
    loader = NpmLoader(package,
                       package_url(package),
                       package_metadata_url(package))

    actual_load_status = loader.load()
    expected_snapshot_id = 'd0587e1195aed5a8800411a008f2f2d627f18e2d'
    assert actual_load_status == {
        'status': 'eventful',
        'snapshot_id': expected_snapshot_id
    }

    stats = get_stats(loader.storage)

    assert {
        'content': len(_expected_new_contents_first_visit),
        'directory': len(_expected_new_directories_first_visit),
        'origin': 1,
        'origin_visit': 1,
        'person': 2,
        'release': 0,
        'revision': len(_expected_new_revisions_first_visit),
        'skipped_content': 0,
        'snapshot': 1,
    } == stats

    assert len(list(loader.storage.content_get(
        _expected_new_contents_first_visit))) == len(
            _expected_new_contents_first_visit)

    assert list(loader.storage.directory_missing(
        _expected_new_directories_first_visit)) == []

    assert list(loader.storage.revision_missing(
        _expected_new_revisions_first_visit)) == []

    expected_snapshot = {
        'id': expected_snapshot_id,
        'branches': {
            'HEAD': {
                'target': 'releases/0.0.4',
                'target_type': 'alias'
            },
            'releases/0.0.2': {
                'target': 'd8a1c7474d2956ac598a19f0f27d52f7015f117e',
                'target_type': 'revision'
            },
            'releases/0.0.3': {
                'target': '5f9eb78af37ffd12949f235e86fac04898f9f72a',
                'target_type': 'revision'
            },
            'releases/0.0.4': {
                'target': 'ba019b192bdb94bd0b5bd68b3a5f92b5acc2239a',
                'target_type': 'revision'
            }
        }
    }
    check_snapshot(expected_snapshot, loader.storage)


def test_npm_loader_incremental_visit(
        swh_config, requests_mock_datadir_visits):
    package = 'org'
    url = package_url(package)
    metadata_url = package_metadata_url(package)
    loader = NpmLoader(package, url, metadata_url)

    actual_load_status = loader.load()
    assert actual_load_status['status'] == 'eventful'
    assert actual_load_status['status'] is not None
    origin_visit = list(loader.storage.origin_visit_get(url))[-1]
    assert origin_visit['status'] == 'full'
    assert origin_visit['type'] == 'npm'

    stats = get_stats(loader.storage)

    assert {
        'content': len(_expected_new_contents_first_visit),
        'directory': len(_expected_new_directories_first_visit),
        'origin': 1,
        'origin_visit': 1,
        'person': 2,
        'release': 0,
        'revision': len(_expected_new_revisions_first_visit),
        'skipped_content': 0,
        'snapshot': 1,
    } == stats

    loader._info = None  # reset loader internal state
    actual_load_status2 = loader.load()
    assert actual_load_status2['status'] == 'eventful'
    snap_id2 = actual_load_status2['snapshot_id']
    assert snap_id2 is not None
    assert snap_id2 != actual_load_status['snapshot_id']

    origin_visit2 = list(loader.storage.origin_visit_get(url))[-1]
    assert origin_visit2['status'] == 'full'
    assert origin_visit2['type'] == 'npm'

    stats = get_stats(loader.storage)

    assert {  # 3 new releases artifacts
        'content': len(_expected_new_contents_first_visit) + 14,
        'directory': len(_expected_new_directories_first_visit) + 15,
        'origin': 1,
        'origin_visit': 2,
        'person': 2,
        'release': 0,
        'revision': len(_expected_new_revisions_first_visit) + 3,
        'skipped_content': 0,
        'snapshot': 2,
    } == stats

    urls = [
        m.url for m in requests_mock_datadir_visits.request_history
        if m.url.startswith('https://registry.npmjs.org')
    ]
    assert len(urls) == len(set(urls))  # we visited each artifact once across



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.model.hashutil import hash_to_bytes

from swh.loader.package.archive import ArchiveLoader, artifact_identity
from swh.loader.package.tests.common import (
    check_snapshot, check_metadata_paths, get_stats
)


URL = 'https://ftp.gnu.org/gnu/8sync/'
GNU_ARTIFACTS = [
    {
        'time': 944729610,
        'url': 'https://ftp.gnu.org/gnu/8sync/8sync-0.1.0.tar.gz',
        'length': 221837,
        'filename': '8sync-0.1.0.tar.gz',
        'version': '0.1.0',
    }
]

_expected_new_contents_first_visit = [
    'e9258d81faf5881a2f96a77ba609396f82cb97ad',
    '1170cf105b04b7e2822a0e09d2acf71da7b9a130',
    'fbd27c3f41f2668624ffc80b7ba5db9b92ff27ac',
    '0057bec9b5422aff9256af240b177ac0e3ac2608',
    '2b8d0d0b43a1078fc708930c8ddc2956a86c566e',
    '27de3b3bc6545d2a797aeeb4657c0e215a0c2e55',
    '2e6db43f5cd764e677f416ff0d0c78c7a82ef19b',
    'ae9be03bd2a06ed8f4f118d3fe76330bb1d77f62',
    'edeb33282b2bffa0e608e9d2fd960fd08093c0ea',
    'd64e64d4c73679323f8d4cde2643331ba6c20af9',
    '7a756602914be889c0a2d3952c710144b3e64cb0',
    '84fb589b554fcb7f32b806951dcf19518d67b08f',
    '8624bcdae55baeef00cd11d5dfcfa60f68710a02',
    'e08441aeab02704cfbd435d6445f7c072f8f524e',
    'f67935bc3a83a67259cda4b2d43373bd56703844',
    '809788434b433eb2e3cfabd5d591c9a659d5e3d8',
    '7d7c6c8c5ebaeff879f61f37083a3854184f6c41',
    'b99fec102eb24bffd53ab61fc30d59e810f116a2',
    '7d149b28eaa228b3871c91f0d5a95a2fa7cb0c68',
    'f0c97052e567948adf03e641301e9983c478ccff',
    '7fb724242e2b62b85ca64190c31dcae5303e19b3',
    '4f9709e64a9134fe8aefb36fd827b84d8b617ab5',
    '7350628ccf194c2c3afba4ac588c33e3f3ac778d',
    '0bb892d9391aa706dc2c3b1906567df43cbe06a2',
    '49d4c0ce1a16601f1e265d446b6c5ea6b512f27c',
    '6b5cc594ac466351450f7f64a0b79fdaf4435ad3',
    '3046e5d1f70297e2a507b98224b6222c9688d610',
    '1572607d456d7f633bc6065a2b3048496d679a31',
]

_expected_new_directories_first_visit = [
    'daabc65ec75d487b1335ffc101c0ac11c803f8fc',
    '263be23b4a8101d3ad0d9831319a3e0f2b065f36',
    '7f6e63ba6eb3e2236f65892cd822041f1a01dd5c',
    '4db0a3ecbc976083e2dac01a62f93729698429a3',
    'dfef1c80e1098dd5deda664bb44a9ab1f738af13',
    'eca971d346ea54d95a6e19d5051f900237fafdaa',
    '3aebc29ed1fccc4a6f2f2010fb8e57882406b528',
]

_expected_new_revisions_first_visit = {
    '44183488c0774ce3c957fa19ba695cf18a4a42b3':
    '3aebc29ed1fccc4a6f2f2010fb8e57882406b528'
}

_expected_branches_first_visit = {
    'HEAD': {
        'target_type': 'alias',
        'target': 'releases/0.1.0',
    },
    'releases/0.1.0': {
        'target_type': 'revision',
        'target': '44183488c0774ce3c957fa19ba695cf18a4a42b3',
    },
}

# hash is different then before as we changed the snapshot
# gnu used to use `release/` (singular) instead of plural
_expected_new_snapshot_first_visit_id = 'c419397fd912039825ebdbea378bc6283f006bf5'  # noqa


def visit_with_no_artifact_found(swh_config, requests_mock_datadir):
    url = URL
    unknown_artifact_url = 'https://ftp.g.o/unknown/8sync-0.1.0.tar.gz'
    loader = ArchiveLoader(url, artifacts=[
        {
            'time': 944729610,
            'url': unknown_artifact_url,  # unknown artifact
            'length': 221837,
            'filename': '8sync-0.1.0.tar.gz',
            'version': '0.1.0',
        }
    ])

    actual_load_status = loader.load()
    assert actual_load_status['status'] == 'uneventful'
    assert actual_load_status['snapshot_id'] is not None
    stats = get_stats(loader.storage)

    assert {
        'content': 0,
        'directory': 0,
        'origin': 1,
        'origin_visit': 1,
        'person': 0,
        'release': 0,
        'revision': 0,
        'skipped_content': 0,
        'snapshot': 1,
    } == stats

    origin_visit = next(loader.storage.origin_visit_get(url))
    assert origin_visit['status'] == 'partial'
    assert origin_visit['type'] == 'tar'


def test_check_revision_metadata_structure(swh_config, requests_mock_datadir):
    loader = ArchiveLoader(url=URL, artifacts=GNU_ARTIFACTS)

    actual_load_status = loader.load()
    assert actual_load_status['status'] == 'eventful'
    assert actual_load_status['snapshot_id'] is not None

    expected_revision_id = hash_to_bytes(
        '44183488c0774ce3c957fa19ba695cf18a4a42b3')
    revision = list(loader.storage.revision_get([expected_revision_id]))[0]

    assert revision is not None

    check_metadata_paths(revision['metadata'], paths=[
        ('intrinsic', dict),
        ('extrinsic.provider', str),
        ('extrinsic.when', str),
        ('extrinsic.raw', dict),
        ('original_artifact', list),
    ])

    for original_artifact in revision['metadata']['original_artifact']:
        check_metadata_paths(original_artifact, paths=[
            ('filename', str),
            ('length', int),
            ('checksums', dict),
        ])


def test_visit_with_release_artifact_no_prior_visit(
        swh_config, requests_mock_datadir):
    """With no prior visit, load a gnu project ends up with 1 snapshot

    """
    loader = ArchiveLoader(url=URL, artifacts=GNU_ARTIFACTS)

    actual_load_status = loader.load()
    assert actual_load_status['status'] == 'eventful'
    assert actual_load_status['snapshot_id'] == _expected_new_snapshot_first_visit_id  # noqa

    stats = get_stats(loader.storage)
    assert {
        'content': len(_expected_new_contents_first_visit),
        'directory': len(_expected_new_directories_first_visit),
        'origin': 1,
        'origin_visit': 1,
        'person': 1,
        'release': 0,
        'revision': len(_expected_new_revisions_first_visit),
        'skipped_content': 0,
        'snapshot': 1
    } == stats

    expected_contents = map(hash_to_bytes, _expected_new_contents_first_visit)
    assert list(loader.storage.content_missing_per_sha1(expected_contents)) \
        == []

    expected_dirs = map(hash_to_bytes, _expected_new_directories_first_visit)
    assert list(loader.storage.directory_missing(expected_dirs)) == []

    expected_revs = map(hash_to_bytes, _expected_new_revisions_first_visit)
    assert list(loader.storage.revision_missing(expected_revs)) == []

    expected_snapshot = {
        'id': _expected_new_snapshot_first_visit_id,
        'branches': _expected_branches_first_visit,
    }

    check_snapshot(expected_snapshot, loader.storage)


def test_2_visits_without_change(swh_config, requests_mock_datadir):
    """With no prior visit, load a gnu project ends up with 1 snapshot

    """
    url = URL
    loader = ArchiveLoader(url, artifacts=GNU_ARTIFACTS)

    actual_load_status = loader.load()
    assert actual_load_status['status'] == 'eventful'
    assert actual_load_status['snapshot_id'] is not None
    origin_visit = list(loader.storage.origin_visit_get(url))[-1]
    assert origin_visit['status'] == 'full'
    assert origin_visit['type'] == 'tar'

    actual_load_status2 = loader.load()
    assert actual_load_status2['status'] == 'uneventful'
    assert actual_load_status2['snapshot_id'] is not None

    assert actual_load_status['snapshot_id'] == actual_load_status2[
        'snapshot_id']

    origin_visit2 = list(loader.storage.origin_visit_get(url))[-1]
    assert origin_visit2['status'] == 'full'
    assert origin_visit2['type'] == 'tar'

    urls = [
        m.url for m in requests_mock_datadir.request_history
        if m.url.startswith('https://ftp.gnu.org')
    ]
    assert len(urls) == 1


def test_2_visits_with_new_artifact(swh_config, requests_mock_datadir):
    """With no prior visit, load a gnu project ends up with 1 snapshot

    """
    url = URL
    artifact1 = GNU_ARTIFACTS[0]
    loader = ArchiveLoader(url, [artifact1])

    actual_load_status = loader.load()
    assert actual_load_status['status'] == 'eventful'
    assert actual_load_status['snapshot_id'] is not None

    origin_visit = list(loader.storage.origin_visit_get(url))[-1]
    assert origin_visit['status'] == 'full'
    assert origin_visit['type'] == 'tar'

    stats = get_stats(loader.storage)
    assert {
        'content': len(_expected_new_contents_first_visit),
        'directory': len(_expected_new_directories_first_visit),
        'origin': 1,
        'origin_visit': 1,
        'person': 1,
        'release': 0,
        'revision': len(_expected_new_revisions_first_visit),
        'skipped_content': 0,
        'snapshot': 1
    } == stats

    urls = [
        m.url for m in requests_mock_datadir.request_history
        if m.url.startswith('https://ftp.gnu.org')
    ]
    assert len(urls) == 1

    artifact2 = {
        'time': 1480991830,
        'url': 'https://ftp.gnu.org/gnu/8sync/8sync-0.2.0.tar.gz',
        'length': 238466,
        'filename': '8sync-0.2.0.tar.gz',
        'version': '0.2.0',
    }

    loader2 = ArchiveLoader(url, [artifact1, artifact2])
    # implementation detail: share the storage in between visits
    loader2.storage = loader.storage
    stats2 = get_stats(loader2.storage)
    assert stats == stats2  # ensure we share the storage

    actual_load_status2 = loader2.load()
    assert actual_load_status2['status'] == 'eventful'
    assert actual_load_status2['snapshot_id'] is not None

    stats2 = get_stats(loader.storage)
    assert {
        'content': len(_expected_new_contents_first_visit) + 14,
        'directory': len(_expected_new_directories_first_visit) + 8,
        'origin': 1,
        'origin_visit': 1 + 1,
        'person': 1,
        'release': 0,
        'revision': len(_expected_new_revisions_first_visit) + 1,
        'skipped_content': 0,
        'snapshot': 1 + 1,
    } == stats2

    origin_visit2 = list(loader.storage.origin_visit_get(url))[-1]
    assert origin_visit2['status'] == 'full'
    assert origin_visit2['type'] == 'tar'

    urls = [
        m.url for m in requests_mock_datadir.request_history
        if m.url.startswith('https://ftp.gnu.org')
    ]
    # 1 artifact (2nd time no modification) + 1 new artifact
    assert len(urls) == 2


def test_artifact_identity():
    """Compute primary key should return the right identity

    """
    data = {
        'a': 1,
        'b': 2,
        'length': 221837,
        'filename': '8sync-0.1.0.tar.gz',
        'version': '0.1.0',
    }

    for id_keys, expected_id in [
            (['a', 'b'], [1, 2]),
            ([], []),
            (['a', 'key-that-does-not-exist'], [1, None])
    ]:
        actual_id = artifact_identity(data, id_keys=id_keys)
        assert actual_id == expected_id


def test_2_visits_without_change_not_gnu(swh_config, requests_mock_datadir):
    """Load a project archive (not gnu) ends up with 1 snapshot

    """
    url = 'https://something.else.org/8sync/'
    artifacts = [  # this is not a gnu artifact
        {
            'time': '1999-12-09T09:53:30+00:00',  # it's also not a timestamp
            'sha256': 'd5d1051e59b2be6f065a9fc6aedd3a391e44d0274b78b9bb4e2b57a09134dbe4',  # noqa
            # keep a gnu artifact reference to avoid adding other test files
            'url': 'https://ftp.gnu.org/gnu/8sync/8sync-0.2.0.tar.gz',
            'length': 238466,
            'filename': '8sync-0.2.0.tar.gz',
            'version': '0.2.0',
        }
    ]

    # Here the loader defines the id_keys to use for existence in the snapshot
    # It's not the default archive loader which
    loader = ArchiveLoader(
        url, artifacts=artifacts, identity_artifact_keys=[
            'sha256', 'length', 'url'])

    actual_load_status = loader.load()
    assert actual_load_status['status'] == 'eventful'
    assert actual_load_status['snapshot_id'] is not None
    origin_visit = list(loader.storage.origin_visit_get(url))[-1]
    assert origin_visit['status'] == 'full'
    assert origin_visit['type'] == 'tar'

    actual_load_status2 = loader.load()
    assert actual_load_status2['status'] == 'uneventful'
    assert actual_load_status2['snapshot_id'] == actual_load_status[
        'snapshot_id']
    origin_visit2 = list(loader.storage.origin_visit_get(url))[-1]
    assert origin_visit2['status'] == 'full'
    assert origin_visit2['type'] == 'tar'

    urls = [
        m.url for m in requests_mock_datadir.request_history
        if m.url.startswith('https://ftp.gnu.org')
    ]
    assert len(urls) == 1



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest
import requests


def test_swh_proxy():
    with pytest.raises(requests.exceptions.ProxyError):
        requests.get('https://www.softwareheritage.org')



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os

from typing import Any, Dict, Generator, Mapping, Optional, Sequence, Tuple
from urllib.parse import urlparse
from pkginfo import UnpackedSDist

import iso8601

from swh.model.identifiers import normalize_timestamp
from swh.loader.package.loader import PackageLoader
from swh.loader.package.utils import api_info, release_name


class PyPILoader(PackageLoader):
    """Load pypi origin's artifact releases into swh archive.

    """
    visit_type = 'pypi'

    def __init__(self, url):
        super().__init__(url=url)
        self._info = None
        self.provider_url = pypi_api_url(self.url)

    @property
    def info(self) -> Dict:
        """Return the project metadata information (fetched from pypi registry)

        """
        if not self._info:
            self._info = api_info(self.provider_url)
        return self._info

    def get_versions(self) -> Sequence[str]:
        return self.info['releases'].keys()

    def get_default_version(self) -> str:
        return self.info['info']['version']

    def get_package_info(self, version: str) -> Generator[
            Tuple[str, Mapping[str, Any]], None, None]:
        res = []
        for meta in self.info['releases'][version]:
            if meta['packagetype'] != 'sdist':
                continue
            filename = meta['filename']
            p_info = {
                'url': meta['url'],
                'filename': filename,
                'raw': meta,
            }
            res.append((version, p_info))

        if len(res) == 1:
            version, p_info = res[0]
            yield release_name(version), p_info
        else:
            for version, p_info in res:
                yield release_name(version, p_info['filename']), p_info

    def resolve_revision_from(
            self, known_artifacts: Dict, artifact_metadata: Dict) \
            -> Optional[bytes]:
        sha256 = artifact_metadata['digests']['sha256']
        for rev_id, known_artifact in known_artifacts.items():
            for original_artifact in known_artifact['original_artifact']:
                if sha256 == original_artifact['checksums']['sha256']:
                    return rev_id
        return None

    def build_revision(
            self, a_metadata: Dict, uncompressed_path: str) -> Dict:
        i_metadata = extract_intrinsic_metadata(uncompressed_path)

        # from intrinsic metadata
        name = i_metadata['version']
        _author = author(i_metadata)

        # from extrinsic metadata
        message = a_metadata.get('comment_text', '')
        message = '%s: %s' % (name, message) if message else name
        date = normalize_timestamp(
            int(iso8601.parse_date(a_metadata['upload_time']).timestamp()))

        return {
            'type': 'tar',
            'message': message.encode('utf-8'),
            'author': _author,
            'date': date,
            'committer': _author,
            'committer_date': date,
            'parents': [],
            'metadata': {
                'intrinsic': {
                    'tool': 'PKG-INFO',
                    'raw': i_metadata,
                },
                'extrinsic': {
                    'provider': self.provider_url,
                    'when': self.visit_date.isoformat(),
                    'raw': a_metadata,
                },
            }
        }


def pypi_api_url(url: str) -> str:
    """Compute api url from a project url

    Args:
        url (str): PyPI instance's url (e.g: https://pypi.org/project/requests)
        This deals with correctly transforming the project's api url (e.g
        https://pypi.org/pypi/requests/json)

    Returns:
        api url

    """
    p_url = urlparse(url)
    project_name = p_url.path.rstrip('/').split('/')[-1]
    url = '%s://%s/pypi/%s/json' % (p_url.scheme, p_url.netloc, project_name)
    return url


def extract_intrinsic_metadata(dir_path: str) -> Dict:
    """Given an uncompressed path holding the pkginfo file, returns a
       pkginfo parsed structure as a dict.

       The release artifact contains at their root one folder. For example:
       $ tar tvf zprint-0.0.6.tar.gz
       drwxr-xr-x root/root         0 2018-08-22 11:01 zprint-0.0.6/
       ...

    Args:

        dir_path (str): Path to the uncompressed directory
                        representing a release artifact from pypi.

    Returns:
        the pkginfo parsed structure as a dict if any or None if
        none was present.

    """
    # Retrieve the root folder of the archive
    if not os.path.exists(dir_path):
        return {}
    lst = os.listdir(dir_path)
    if len(lst) != 1:
        return {}
    project_dirname = lst[0]
    pkginfo_path = os.path.join(dir_path, project_dirname, 'PKG-INFO')
    if not os.path.exists(pkginfo_path):
        return {}
    pkginfo = UnpackedSDist(pkginfo_path)
    raw = pkginfo.__dict__
    raw.pop('filename')  # this gets added with the ondisk location
    return raw


def author(data: Dict) -> Dict:
    """Given a dict of project/release artifact information (coming from
       PyPI), returns an author subset.

    Args:
        data (dict): Representing either artifact information or
                     release information.

    Returns:
        swh-model dict representing a person.

    """
    name = data.get('author')
    email = data.get('author_email')
    fullname = None  # type: Optional[str]

    if email:
        fullname = '%s <%s>' % (name, email)
    else:
        fullname = name

    if not fullname:
        return {'fullname': b'', 'name': None, 'email': None}

    if name is not None:
        name = name.encode('utf-8')

    if email is not None:
        email = email.encode('utf-8')

    return {
        'fullname': fullname.encode('utf-8'),
        'name': name,
        'email': email
    }



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import pytest
import shutil
import subprocess
import tempfile

from unittest import TestCase

from swh.model import hashutil
from swh.model.hashutil import hash_to_bytes


class BaseLoaderStorageTest:
    def _assertCountEqual(self, type, expected_length, msg=None):
        """Check typed 'type' state to have the same expected length.

        """
        self.storage.refresh_stat_counters()
        self.assertEqual(self.storage.stat_counters()[type],
                         expected_length, msg=msg)

    def assertCountContents(self, len_expected_contents, msg=None):
        self._assertCountEqual('content', len_expected_contents, msg=msg)

    def assertCountDirectories(self, len_expected_directories, msg=None):
        self._assertCountEqual('directory', len_expected_directories,
                               msg=msg)

    def assertCountReleases(self, len_expected_releases, msg=None):
        self._assertCountEqual('release', len_expected_releases, msg=msg)

    def assertCountRevisions(self, len_expected_revisions, msg=None):
        self._assertCountEqual('revision', len_expected_revisions, msg=msg)

    def assertCountSnapshots(self, len_expected_snapshot, msg=None):
        self._assertCountEqual('snapshot', len_expected_snapshot, msg=msg)

    def assertContentsContain(self, expected_contents):
        """Check the provided content are a subset of the stored ones.

        Args:
            expected_contents ([sha1]): List of content ids"""
        missing = list(self.storage.content_missing(
            {'sha1': hash_to_bytes(content_hash)}
            for content_hash in expected_contents))
        self.assertEqual(missing, [])

    def assertDirectoriesContain(self, expected_directories):
        """Check the provided directories are a subset of the stored ones.

        Args:
            expected_directories ([sha1]): List of directory ids."""
        missing = list(self.storage.directory_missing(
            hash_to_bytes(dir_) for dir_ in expected_directories))
        self.assertEqual(missing, [])

    def assertReleasesContain(self, expected_releases):
        """Check the provided releases are a subset of the stored ones.

        Args:
            releases (list): list of swh releases' identifiers.

        """
        missing = list(self.storage.release_missing(
            hash_to_bytes(rel) for rel in expected_releases))
        self.assertEqual(missing, [])

    def assertRevisionsContain(self, expected_revisions):
        """Check the provided revisions are a subset of the stored ones.

        Expects self.loader to be instantiated and ready to be
        inspected (meaning the loading took place).

        Args:
            expected_revisions (dict): Dict with key revision id,
            value the targeted directory id.

        """
        revs = list(self.storage.revision_get(
            hashutil.hash_to_bytes(rev_id) for rev_id in expected_revisions))
        self.assertNotIn(None, revs)
        self.assertEqual(
            {rev['id']: rev['directory'] for rev in revs},
            {hash_to_bytes(rev_id): hash_to_bytes(rev_dir)
             for (rev_id, rev_dir) in expected_revisions.items()})

    def assertSnapshotEqual(self, expected_snapshot, expected_branches=[]):
        """Check for snapshot match.

        Provide the hashes as hexadecimal, the conversion is done
        within the method.

        Args:

            expected_snapshot (str/dict): Either the snapshot
                                          identifier or the full
                                          snapshot
            expected_branches (dict): expected branches or nothing is
                                      the full snapshot is provided

        """
        if isinstance(expected_snapshot, dict) and not expected_branches:
            expected_snapshot_id = expected_snapshot['id']
            expected_branches = expected_snapshot['branches']
        else:
            expected_snapshot_id = expected_snapshot

        snap = self.storage.snapshot_get(hash_to_bytes(expected_snapshot_id))
        self.assertIsNotNone(snap)

        def decode_target(target):
            if not target:
                return target
            target_type = target['target_type']

            if target_type == 'alias':
                decoded_target = target['target'].decode('utf-8')
            else:
                decoded_target = hashutil.hash_to_hex(target['target'])

            return {
                'target': decoded_target,
                'target_type': target_type
            }

        branches = {
            branch.decode('utf-8'): decode_target(target)
            for branch, target in snap['branches'].items()
        }
        self.assertEqual(expected_branches, branches)

    def assertOriginMetadataContains(self, origin_url,
                                     expected_origin_metadata):
        """Check the storage contains this metadata for the given origin.

        Args:

            origin_url (str): URL of the origin
            expected_origin_metadata (dict):
                              Extrinsic metadata of the origin
                              <https://forge.softwareheritage.org/T1344>
        """
        origin = self.storage.origin_get({'url': origin_url})
        results = self.storage.origin_metadata_get_by(origin['url'])
        self.assertEqual(len(results), 1, results)
        result = results[0]
        self.assertEqual(result['metadata'], expected_origin_metadata)


@pytest.mark.fs
class BaseLoaderTest(TestCase, BaseLoaderStorageTest):
    """Mixin base loader test class.

    This allows to uncompress archives (mercurial, svn, git,
    ... repositories) into a temporary folder so that the loader under
    test can work with this.

    When setUp() is done, the following variables are defined:
    - self.repo_url: can be used as an origin_url for example
    - self.destination_path: can be used as a path to ingest the
                             <techno> repository.

    Args:
        archive_name (str): Name of the archive holding the repository
                            (folder, repository, dump, etc...)
        start_path (str): (mandatory) Path from where starting to look
                                      for resources
        filename (Optional[str]): Name of the filename/folder once the
            archive is uncompressed. When the filename is not
            provided, the archive name is used as a derivative. This
            is used both for the self.repo_url and
            self.destination_path computation (this one only when
            provided)
        resources_path (str): Folder name to look for archive
        prefix_tmp_folder_name (str): Prefix name to name the temporary folder
        uncompress_archive (bool): Uncompress the archive passed as
                                  parameters (default to True). It so
                                  happens we could avoid doing
                                  anything to the tarball.

    """
    def setUp(self, archive_name, *, start_path, filename=None,
              resources_path='resources', prefix_tmp_folder_name='',
              uncompress_archive=True):
        super().setUp()
        repo_path = os.path.join(start_path, resources_path, archive_name)
        if not uncompress_archive:
            # In that case, simply sets the archive's path
            self.destination_path = repo_path
            self.tmp_root_path = None
            self.repo_url = 'file://' + repo_path
            return
        tmp_root_path = tempfile.mkdtemp(
            prefix=prefix_tmp_folder_name, suffix='-tests')
        # uncompress folder/repositories/dump for the loader to ingest
        subprocess.check_output(['tar', 'xf', repo_path, '-C', tmp_root_path])
        # build the origin url (or some derivative form)
        _fname = filename if filename else os.path.basename(archive_name)
        self.repo_url = 'file://' + tmp_root_path + '/' + _fname
        # where is the data to ingest?
        if filename:
            # archive holds one folder with name <filename>
            self.destination_path = os.path.join(tmp_root_path, filename)
        else:
            self.destination_path = tmp_root_path
        self.tmp_root_path = tmp_root_path

    def tearDown(self):
        """Clean up temporary working directory

        """
        if self.tmp_root_path and os.path.exists(self.tmp_root_path):
            shutil.rmtree(self.tmp_root_path)



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import iso8601
import logging

from os import path
from typing import Any, Dict, Generator, Mapping, Optional, Sequence, Tuple

from swh.loader.package.loader import PackageLoader
from swh.loader.package.utils import release_name
from swh.model.identifiers import normalize_timestamp


logger = logging.getLogger(__name__)
SWH_PERSON = {
    'name': b'Software Heritage',
    'fullname': b'Software Heritage',
    'email': b'robot@softwareheritage.org'
}
REVISION_MESSAGE = b'swh-loader-package: synthetic revision message'


class ArchiveLoader(PackageLoader):
    visit_type = 'tar'

    def __init__(self, url: str, artifacts: Sequence[Mapping[str, Any]],
                 identity_artifact_keys: Optional[Sequence[str]] = None):
        """Loader constructor.

        For now, this is the lister's task output.

        Args:
            url: Origin url
            artifacts: List of artifact information with keys:

               **time**: last modification time as either isoformat date string
                   or timestamp
               **url**: the artifact url to retrieve filename
               **artifact's filename version**: artifact's version length
               **length**: artifact's length

            identity_artifact_keys: Optional List of keys forming the
                "identity" of an artifact

        """
        super().__init__(url=url)
        self.artifacts = artifacts  # assume order is enforced in the lister
        if not identity_artifact_keys:
            # default keys for gnu
            identity_artifact_keys = ['time', 'url', 'length', 'version']
        self.identity_artifact_keys = identity_artifact_keys

    def get_versions(self) -> Sequence[str]:
        versions = []
        for archive in self.artifacts:
            v = archive.get('version')
            if v:
                versions.append(v)
        return versions

    def get_default_version(self) -> str:
        # It's the most recent, so for this loader, it's the last one
        return self.artifacts[-1]['version']

    def get_package_info(self, version: str) -> Generator[
            Tuple[str, Mapping[str, Any]], None, None]:
        for a_metadata in self.artifacts:
            url = a_metadata['url']
            package_version = a_metadata['version']
            if version == package_version:
                filename = a_metadata.get('filename')
                p_info = {
                    'url': url,
                    'filename': filename if filename else path.split(url)[-1],
                    'raw': a_metadata,
                }
                # FIXME: this code assumes we have only 1 artifact per
                # versioned package
                yield release_name(version), p_info

    def resolve_revision_from(
            self, known_artifacts: Dict, artifact_metadata: Dict) \
            -> Optional[bytes]:
        identity = artifact_identity(
            artifact_metadata, id_keys=self.identity_artifact_keys)
        for rev_id, known_artifact in known_artifacts.items():
            logging.debug('known_artifact: %s', known_artifact)
            reference_artifact = known_artifact['extrinsic']['raw']
            known_identity = artifact_identity(
                reference_artifact, id_keys=self.identity_artifact_keys)
            if identity == known_identity:
                return rev_id
        return None

    def build_revision(self, a_metadata: Mapping[str, Any],
                       uncompressed_path: str) -> Dict:
        time = a_metadata['time']  # assume it's a timestamp
        if isinstance(time, str):  # otherwise, assume it's a parsable date
            time = iso8601.parse_date(time)
        normalized_time = normalize_timestamp(time)
        return {
            'type': 'tar',
            'message': REVISION_MESSAGE,
            'date': normalized_time,
            'author': SWH_PERSON,
            'committer': SWH_PERSON,
            'committer_date': normalized_time,
            'parents': [],
            'metadata': {
                'intrinsic': {},
                'extrinsic': {
                    'provider': self.url,
                    'when': self.visit_date.isoformat(),
                    'raw': a_metadata,
                },
            },
        }


def artifact_identity(d: Mapping[str, Any],
                      id_keys: Sequence[str]) -> Sequence[Any]:
    """Compute the primary key for a dict using the id_keys as primary key
       composite.

    Args:
        d: A dict entry to compute the primary key on
        id_keys: Sequence of keys to use as primary key

    Returns:
        The identity for that dict entry

    """
    return [d.get(k) for k in id_keys]



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import requests

from typing import (
    Any, Dict, Generator, List, Mapping, Optional, Sequence, Tuple, Union
)

from swh.model.hashutil import hash_to_hex, hash_to_bytes
from swh.loader.package.loader import PackageLoader
from swh.loader.package.utils import download


logger = logging.getLogger(__name__)


class DepositLoader(PackageLoader):
    """Load pypi origin's artifact releases into swh archive.

    """
    visit_type = 'deposit'

    def __init__(self, url: str, deposit_id: str):
        """Constructor

        Args:
            url: Origin url to associate the artifacts/metadata to
            deposit_id: Deposit identity

        """
        super().__init__(url=url)

        config_deposit = self.config['deposit']
        self.deposit_id = deposit_id
        self.client = ApiClient(url=config_deposit['url'],
                                auth=config_deposit['auth'])
        self._metadata = None

    @property
    def metadata(self):
        if self._metadata is None:
            self._metadata = self.client.metadata_get(self.deposit_id)
        return self._metadata

    def get_versions(self) -> Sequence[str]:
        # only 1 branch 'HEAD' with no alias since we only have 1 snapshot
        # branch
        return ['HEAD']

    def get_package_info(self, version: str) -> Generator[
            Tuple[str, Mapping[str, Any]], None, None]:
        p_info = {
            'filename': 'archive.zip',
            'raw': self.metadata,
        }
        yield 'HEAD', p_info

    def download_package(self, p_info: Mapping[str, Any],
                         tmpdir: str) -> List[Tuple[str, Mapping]]:
        """Override to allow use of the dedicated deposit client

        """
        return [self.client.archive_get(
            self.deposit_id, tmpdir, p_info['filename'])]

    def build_revision(
            self, a_metadata: Dict, uncompressed_path: str) -> Dict:
        revision = a_metadata.pop('revision')
        metadata = {
            'extrinsic': {
                'provider': self.client.metadata_url(self.deposit_id),
                'when': self.visit_date.isoformat(),
                'raw': a_metadata,
            },
        }

        # FIXME: the deposit no longer needs to build the revision
        revision['metadata'].update(metadata)
        revision['author'] = parse_author(revision['author'])
        revision['committer'] = parse_author(revision['committer'])
        revision['message'] = revision['message'].encode('utf-8')
        revision['type'] = 'tar'

        return revision

    def load(self) -> Dict:
        # Usual loading
        r = super().load()
        success = r['status'] != 'failed'

        if success:
            # Update archive with metadata information
            origin_metadata = self.metadata['origin_metadata']

            logger.debug('origin_metadata: %s', origin_metadata)
            tools = self.storage.tool_add([origin_metadata['tool']])
            logger.debug('tools: %s', tools)
            tool_id = tools[0]['id']

            provider = origin_metadata['provider']
            # FIXME: Shall we delete this info?
            provider_id = self.storage.metadata_provider_add(
                provider['provider_name'],
                provider['provider_type'],
                provider['provider_url'],
                metadata=None)

            metadata = origin_metadata['metadata']
            self.storage.origin_metadata_add(
                self.url, self.visit_date, provider_id, tool_id, metadata)

        # Update deposit status
        try:
            if not success:
                self.client.status_update(self.deposit_id, status='failed')
                return r

            snapshot_id = hash_to_bytes(r['snapshot_id'])
            branches = self.storage.snapshot_get(snapshot_id)['branches']
            logger.debug('branches: %s', branches)
            if not branches:
                return r
            rev_id = branches[b'HEAD']['target']

            revision = next(self.storage.revision_get([rev_id]))

            # Retrieve the revision identifier
            dir_id = revision['directory']

            # update the deposit's status to success with its
            # revision-id and directory-id
            self.client.status_update(
                self.deposit_id,
                status='done',
                revision_id=hash_to_hex(rev_id),
                directory_id=hash_to_hex(dir_id),
                origin_url=self.url)
        except Exception:
            logger.exception(
                'Problem when trying to update the deposit\'s status')
            return {'status': 'failed'}
        return r


def parse_author(author):
    """See prior fixme

    """
    return {
        'fullname': author['fullname'].encode('utf-8'),
        'name': author['name'].encode('utf-8'),
        'email': author['email'].encode('utf-8'),
    }


class ApiClient:
    """Private Deposit Api client

    """
    def __init__(self, url, auth: Optional[Mapping[str, str]]):
        self.base_url = url.rstrip('/')
        self.auth = None if not auth else (auth['username'], auth['password'])

    def do(self, method: str, url: str, *args, **kwargs):
        """Internal method to deal with requests, possibly with basic http
           authentication.

        Args:
            method (str): supported http methods as in get/post/put

        Returns:
            The request's execution output

        """
        method_fn = getattr(requests, method)
        if self.auth:
            kwargs['auth'] = self.auth
        return method_fn(url, *args, **kwargs)

    def archive_get(
            self, deposit_id: Union[int, str], tmpdir: str,
            filename: str) -> Tuple[str, Dict]:
        """Retrieve deposit's archive artifact locally

        """
        url = f'{self.base_url}/{deposit_id}/raw/'
        return download(url, dest=tmpdir, filename=filename, auth=self.auth)

    def metadata_url(self, deposit_id: Union[int, str]) -> str:
        return f'{self.base_url}/{deposit_id}/meta/'

    def metadata_get(self, deposit_id: Union[int, str]) -> Dict[str, Any]:
        """Retrieve deposit's metadata artifact as json

        """
        url = self.metadata_url(deposit_id)
        r = self.do('get', url)
        if r.ok:
            return r.json()

        msg = f'Problem when retrieving deposit metadata at {url}'
        logger.error(msg)
        raise ValueError(msg)

    def status_update(self, deposit_id: Union[int, str], status: str,
                      revision_id: Optional[str] = None,
                      directory_id: Optional[str] = None,
                      origin_url: Optional[str] = None):
        """Update deposit's information including status, and persistent
           identifiers result of the loading.

        """
        url = f'/{self.base_url}/{deposit_id}/update/'
        payload = {'status': status}
        if revision_id:
            payload['revision_id'] = revision_id
        if directory_id:
            payload['directory_id'] = directory_id
        if origin_url:
            payload['origin_url'] = origin_url

        self.do('put', url, json=payload)



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from os import path

import logging

from typing import Dict, List, Tuple

from swh.model.hashutil import hash_to_bytes, hash_to_hex


logger = logging.getLogger(__file__)


DATADIR = path.join(path.abspath(path.dirname(__file__)), 'resources')


def decode_target(target):
    """Test helper to ease readability in test

    """
    if not target:
        return target
    target_type = target['target_type']

    if target_type == 'alias':
        decoded_target = target['target'].decode('utf-8')
    else:
        decoded_target = hash_to_hex(target['target'])

    return {
        'target': decoded_target,
        'target_type': target_type
    }


def check_snapshot(expected_snapshot, storage):
    """Check for snapshot match.

    Provide the hashes as hexadecimal, the conversion is done
    within the method.

    Args:
        expected_snapshot (dict): full snapshot with hex ids
        storage (Storage): expected storage

    """
    expected_snapshot_id = expected_snapshot['id']
    expected_branches = expected_snapshot['branches']
    snap = storage.snapshot_get(hash_to_bytes(expected_snapshot_id))
    if snap is None:
        # display known snapshots instead if possible
        if hasattr(storage, '_snapshots'):  # in-mem storage
            from pprint import pprint
            for snap_id, (_snap, _) in storage._snapshots.items():
                snapd = _snap.to_dict()
                snapd['id'] = hash_to_hex(snapd['id'])
                branches = {
                    branch.decode('utf-8'): decode_target(target)
                    for branch, target in snapd['branches'].items()
                }
                snapd['branches'] = branches
                pprint(snapd)
        raise AssertionError('Snapshot is not found')

    branches = {
        branch.decode('utf-8'): decode_target(target)
        for branch, target in snap['branches'].items()
    }
    assert expected_branches == branches


def check_metadata(metadata: Dict, key_path: str, raw_type: str):
    """Given a metadata dict, ensure the associated key_path value is of type
       raw_type.

    Args:
        metadata: Dict to check
        key_path: Path to check
        raw_type: Type to check the path with

    Raises:
        Assertion error in case of mismatch

    """
    data = metadata
    keys = key_path.split('.')
    for k in keys:
        try:
            data = data[k]
        except (TypeError, KeyError) as e:
            # KeyError: because path too long
            # TypeError: data is not a dict
            raise AssertionError(e)
    assert isinstance(data, raw_type)  # type: ignore


def check_metadata_paths(metadata: Dict, paths: List[Tuple[str, str]]):
    """Given a metadata dict, ensure the keys are of expected types

    Args:
        metadata: Dict to check
        key_path: Path to check
        raw_type: Type to check the path with

    Raises:
        Assertion error in case of mismatch

    """
    for key_path, raw_type in paths:
        check_metadata(metadata, key_path, raw_type)


def get_stats(storage) -> Dict:
    """Adaptation utils to unify the stats counters across storage
       implementation.

    """
    storage.refresh_stat_counters()
    stats = storage.stat_counters()

    keys = ['content', 'directory', 'origin', 'origin_visit', 'person',
            'release', 'revision', 'skipped_content', 'snapshot']
    return {k: stats.get(k) for k in keys}



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.model.hashutil import hash_to_bytes
from swh.loader.package.tests.common import (
    decode_target, check_snapshot, check_metadata, check_metadata_paths
)
from swh.storage import get_storage

hash_hex = '43e45d56f88993aae6a0198013efa80716fd8920'


def test_decode_target_edge():
    assert not decode_target(None)


def test_decode_target():
    actual_alias_decode_target = decode_target({
        'target_type': 'alias',
        'target': b'something',
    })

    assert actual_alias_decode_target == {
        'target_type': 'alias',
        'target': 'something',
    }

    actual_decode_target = decode_target({
        'target_type': 'revision',
        'target': hash_to_bytes(hash_hex),
    })

    assert actual_decode_target == {
        'target_type': 'revision',
        'target': hash_hex,
    }


def test_check_snapshot():
    storage = get_storage(cls='memory', args={})

    snap_id = '2498dbf535f882bc7f9a18fb16c9ad27fda7bab7'
    snapshot = {
        'id': hash_to_bytes(snap_id),
        'branches': {
            b'master': {
                'target': hash_to_bytes(hash_hex),
                'target_type': 'revision',
            },
        },
    }

    s = storage.snapshot_add([snapshot])
    assert s == {
        'snapshot:add': 1,
    }

    expected_snapshot = {
        'id': snap_id,
        'branches': {
            'master': {
                'target': hash_hex,
                'target_type': 'revision',
            }
        }
    }
    check_snapshot(expected_snapshot, storage)


def test_check_snapshot_failure():
    storage = get_storage(cls='memory', args={})

    snapshot = {
        'id': hash_to_bytes('2498dbf535f882bc7f9a18fb16c9ad27fda7bab7'),
        'branches': {
            b'master': {
                'target': hash_to_bytes(hash_hex),
                'target_type': 'revision',
            },
        },
    }

    s = storage.snapshot_add([snapshot])
    assert s == {
        'snapshot:add': 1,
    }

    unexpected_snapshot = {
        'id': '2498dbf535f882bc7f9a18fb16c9ad27fda7bab7',
        'branches': {
            'master': {
                'target': hash_hex,
                'target_type': 'release',  # wrong value
            }
        }
    }

    with pytest.raises(AssertionError):
        check_snapshot(unexpected_snapshot, storage)


def test_check_metadata():
    metadata = {
        'a': {
            'raw': {
                'time': 'something',
            },
        },
        'b': [],
        'c': 1,
    }

    for raw_path, raw_type in [
        ('a.raw', dict),
        ('a.raw.time', str),
        ('b', list),
        ('c', int),
    ]:
        check_metadata(metadata, raw_path, raw_type)


def test_check_metadata_ko():
    metadata = {
        'a': {
            'raw': 'hello',
        },
        'b': [],
        'c': 1,
    }

    for raw_path, raw_type in [
        ('a.b', dict),
        ('a.raw.time', str),
    ]:
        with pytest.raises(AssertionError):
            check_metadata(metadata, raw_path, raw_type)


def test_check_metadata_paths():
    metadata = {
        'a': {
            'raw': {
                'time': 'something',
            },
        },
        'b': [],
        'c': 1,
    }

    check_metadata_paths(metadata, [
        ('a.raw', dict),
        ('a.raw.time', str),
        ('b', list),
        ('c', int),
    ])


def test_check_metadata_paths_ko():
    metadata = {
        'a': {
            'raw': 'hello',
        },
        'b': [],
        'c': 1,
    }

    with pytest.raises(AssertionError):
        check_metadata_paths(metadata, [
            ('a.b', dict),
            ('a.raw.time', str),
        ])



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os.path
from os.path import join
import pytest
import re
import tarfile
import tempfile
from unittest import TestCase

from . import BaseLoaderStorageTest

from swh.loader.core.tests import BaseLoaderTest


class TestsTest(BaseLoaderStorageTest, TestCase):
    """Test the helpers provided to other loaders' tests."""

    def _build_archive(self, fd):
        with tarfile.open(mode='w', fileobj=fd) as tf:
            with tempfile.TemporaryDirectory() as src_dir:
                with open(join(src_dir, 'hello.txt'), 'a') as src_file:
                    src_file.write('world\n')
                tf.add(src_dir, arcname='test_dir')

    def _build_workdir(self, workdir):
        os.mkdir(join(workdir, 'resources'))
        tarball_path = join(workdir, 'resources', 'test_archive.tar')
        with open(tarball_path, 'a+b') as tar_fd:
            self._build_archive(tar_fd)

    @pytest.mark.fs
    def test_uncompress_setup_auto_name(self):
        loader_test = BaseLoaderTest()
        with tempfile.TemporaryDirectory() as workdir:
            self._build_workdir(workdir)

            loader_test.setUp('test_archive.tar', start_path=workdir)

        self.assertTrue(re.match('^file://.*-tests/test_archive.tar$',
                                 loader_test.repo_url),
                        msg=loader_test.repo_url)
        self.assertTrue(os.path.isdir(loader_test.destination_path))
        self.assertTrue(os.path.isdir(join(loader_test.destination_path,
                                           'test_dir')))
        self.assertTrue(os.path.isfile(join(loader_test.destination_path,
                                            'test_dir',
                                            'hello.txt')))

        loader_test.tearDown()
        self.assertFalse(os.path.isdir(loader_test.destination_path))

    @pytest.mark.fs
    def test_uncompress_setup_provided_name(self):
        loader_test = BaseLoaderTest()
        with tempfile.TemporaryDirectory() as workdir:
            self._build_workdir(workdir)

            loader_test.setUp('test_archive.tar', start_path=workdir,
                              filename='test_dir')

        self.assertTrue(re.match('^file://.*-tests/test_dir$',
                                 loader_test.repo_url),
                        msg=loader_test.repo_url)
        self.assertTrue(os.path.isdir(loader_test.destination_path))
        self.assertTrue(os.path.isfile(join(loader_test.destination_path,
                                            'hello.txt')))

        loader_test.tearDown()
        self.assertFalse(os.path.isdir(loader_test.destination_path))

    @pytest.mark.fs
    def test_setup_no_uncompress(self):
        loader_test = BaseLoaderTest()
        with tempfile.TemporaryDirectory() as workdir:
            self._build_workdir(workdir)

            loader_test.setUp('test_archive.tar', start_path=workdir,
                              uncompress_archive=False)

        self.assertEqual('file://' + workdir + '/resources/test_archive.tar',
                         loader_test.repo_url)
        self.assertEqual(workdir + '/resources/test_archive.tar',
                         loader_test.destination_path)



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import pytest
import yaml

from typing import Any, Dict

from swh.storage.tests.conftest import * # noqa
from swh.scheduler.tests.conftest import *  # noqa


@pytest.fixture
def swh_loader_config(swh_storage_postgresql) -> Dict[str, Any]:
    return {
        'storage': {
            'cls': 'local',
            'args': {
                'db': swh_storage_postgresql.dsn,
                'objstorage': {
                    'cls': 'memory',
                    'args': {}
                },
            },
        },
        'deposit': {
            'url': 'https://deposit.softwareheritage.org/1/private',
            'auth': {
                'username': 'user',
                'password': 'pass',
            }
        },
    }


@pytest.fixture
def swh_config(swh_loader_config, monkeypatch, tmp_path):
    conffile = os.path.join(str(tmp_path), 'loader.yml')
    with open(conffile, 'w') as f:
        f.write(yaml.dump(swh_loader_config))
    monkeypatch.setenv('SWH_CONFIG_FILENAME', conffile)
    return conffile


@pytest.fixture(autouse=True, scope='session')
def swh_proxy():
    """Automatically inject this fixture in all tests to ensure no outside
       connection takes place.

    """
    os.environ['http_proxy'] = 'http://localhost:999'
    os.environ['https_proxy'] = 'http://localhost:999'


@pytest.fixture(scope='session')  # type: ignore  # expected redefinition
def celery_includes():
    return [
        'swh.loader.package.tasks',
    ]



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
import logging
import os
import requests

from typing import Dict, Optional, Tuple

from swh.model.hashutil import MultiHash, HASH_BLOCK_SIZE
from swh.loader.package import DEFAULT_PARAMS


logger = logging.getLogger(__name__)


def api_info(url: str) -> Dict:
    """Basic api client to retrieve information on project. This deals with
       fetching json metadata about pypi projects.

    Args:
        url (str): The api url (e.g PyPI, npm, etc...)

    Raises:
        ValueError in case of query failures (for some reasons: 404, ...)

    Returns:
        The associated response's information dict

    """
    response = requests.get(url, **DEFAULT_PARAMS)
    if response.status_code != 200:
        raise ValueError("Fail to query '%s'. Reason: %s" % (
            url, response.status_code))
    return response.json()


def download(url: str, dest: str, hashes: Dict = {},
             filename: Optional[str] = None,
             auth: Optional[Tuple[str, str]] = None) -> Tuple[str, Dict]:
    """Download a remote tarball from url, uncompresses and computes swh hashes
       on it.

    Args:
        url: Artifact uri to fetch, uncompress and hash
        dest: Directory to write the archive to
        hashes: Dict of expected hashes (key is the hash algo) for the artifact
            to download (those hashes are expected to be hex string)
        auth: Optional tuple of login/password (for http authentication
            service, e.g. deposit)

    Raises:
        ValueError in case of any error when fetching/computing (length,
        checksums mismatched...)

    Returns:
        Tuple of local (filepath, hashes of filepath)

    """
    params = copy.deepcopy(DEFAULT_PARAMS)
    if auth is not None:
        params['auth'] = auth
    response = requests.get(url, **params, stream=True)
    logger.debug('headers: %s', response.headers)
    if response.status_code != 200:
        raise ValueError("Fail to query '%s'. Reason: %s" % (
            url, response.status_code))
    length = int(response.headers['content-length'])

    filename = filename if filename else os.path.basename(url)
    logger.debug('filename: %s', filename)
    filepath = os.path.join(dest, filename)
    logger.debug('filepath: %s', filepath)

    h = MultiHash(length=length)
    with open(filepath, 'wb') as f:
        for chunk in response.iter_content(chunk_size=HASH_BLOCK_SIZE):
            h.update(chunk)
            f.write(chunk)

    actual_length = os.path.getsize(filepath)
    if length != actual_length:
        raise ValueError('Error when checking size: %s != %s' % (
            length, actual_length))

    # Also check the expected hashes if provided
    if hashes:
        actual_hashes = h.hexdigest()
        for algo_hash in hashes.keys():
            actual_digest = actual_hashes[algo_hash]
            expected_digest = hashes[algo_hash]
            if actual_digest != expected_digest:
                raise ValueError(
                    'Failure when fetching %s. '
                    'Checksum mismatched: %s != %s' % (
                        url, expected_digest, actual_digest))

    extrinsic_metadata = {
        'length': length,
        'filename': filename,
        'checksums': {
            **h.hexdigest()
        },
    }

    logger.debug('extrinsic_metadata', extrinsic_metadata)

    return filepath, extrinsic_metadata


def release_name(version: str, filename: Optional[str] = None) -> str:
    if filename:
        return 'releases/%s/%s' % (version, filename)
    return 'releases/%s' % version



# Copyright (C) 2015-2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import tempfile
import unittest
from unittest.mock import Mock

from swh.loader.core import converters
from swh.model.from_disk import Content


def tmpfile_with_content(fromdir, contentfile):
    """Create a temporary file with content contentfile in directory fromdir.

    """
    tmpfilepath = tempfile.mktemp(
        suffix='.swh',
        prefix='tmp-file-for-test',
        dir=fromdir)

    with open(tmpfilepath, 'wb') as f:
        f.write(contentfile)

    return tmpfilepath


class TestContentForStorage(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        super().setUpClass()
        self.tmpdir = tempfile.TemporaryDirectory(
            prefix='test-swh-loader-core.'
        )

    def tearDown(self):
        self.tmpdir.cleanup()

    def test_content_for_storage_path(self):
        # given
        data = b'temp file for testing content storage conversion'
        tmpfile = tmpfile_with_content(self.tmpdir.name, data)

        obj = Content.from_file(path=os.fsdecode(tmpfile),
                                save_path=True).get_data()

        expected_content = obj.copy()
        expected_content['data'] = data
        expected_content['status'] = 'visible'
        del expected_content['path']
        del expected_content['perms']

        # when
        content = converters.content_for_storage(obj)

        # then
        self.assertEqual(content, expected_content)

    def test_content_for_storage_data(self):
        # given
        data = b'temp file for testing content storage conversion'

        obj = Content.from_bytes(data=data, mode=0o100644).get_data()

        expected_content = obj.copy()
        expected_content['status'] = 'visible'
        del expected_content['perms']

        # when
        content = converters.content_for_storage(obj)

        # then
        self.assertEqual(content, expected_content)

    def test_content_for_storage_too_long(self):
        # given
        data = b'temp file for testing content storage conversion'

        obj = Content.from_bytes(data=data, mode=0o100644).get_data()

        log = Mock()

        expected_content = obj.copy()
        expected_content.pop('data')
        expected_content['status'] = 'absent'
        expected_content['origin'] = 'http://example.org/'
        expected_content['reason'] = 'Content too large'

        # when
        content = converters.content_for_storage(
            obj, log, max_content_size=len(data) - 1,
            origin_url=expected_content['origin'],
        )

        # then
        self.assertEqual(content, expected_content)
        self.assertTrue(log.info.called)
        self.assertIn('Skipping content', log.info.call_args[0][0])
        self.assertIn('too large', log.info.call_args[0][0])



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from typing import Any, Dict, Mapping

import pkg_resources


try:
    __version__ = pkg_resources.get_distribution('swh.loader.core').version
except pkg_resources.DistributionNotFound:
    __version__ = 'devel'


DEFAULT_PARAMS: Dict[str, Any] = {
    'headers': {
        'User-Agent': 'Software Heritage Loader (%s)' % (
            __version__
        )
    }
}


def register() -> Mapping[str, Any]:
    return {
        'task_modules': ['%s.tasks' % __name__],
    }



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import re


from swh.model.hashutil import hash_to_bytes
from swh.loader.package.deposit import DepositLoader

from swh.loader.package.tests.common import (
    check_snapshot, check_metadata_paths, get_stats
)

from swh.core.pytest_plugin import requests_mock_datadir_factory


def test_deposit_init_ok(swh_config, swh_loader_config):
    url = 'some-url'
    deposit_id = 999
    loader = DepositLoader(url, deposit_id)  # Something that does not exist

    assert loader.url == url
    assert loader.client is not None
    assert loader.client.base_url == swh_loader_config['deposit']['url']


def test_deposit_loading_failure_to_fetch_metadata(swh_config):
    """Error during fetching artifact ends us with failed/partial visit

    """
    # private api url form: 'https://deposit.s.o/1/private/hal/666/raw/'
    url = 'some-url'
    unknown_deposit_id = 666
    loader = DepositLoader(url, unknown_deposit_id)  # does not exist

    actual_load_status = loader.load()
    assert actual_load_status == {'status': 'failed'}

    stats = get_stats(loader.storage)

    assert {
        'content': 0,
        'directory': 0,
        'origin': 1,
        'origin_visit': 1,
        'person': 0,
        'release': 0,
        'revision': 0,
        'skipped_content': 0,
        'snapshot': 0,
    } == stats

    origin_visit = next(loader.storage.origin_visit_get(url))
    assert origin_visit['status'] == 'partial'
    assert origin_visit['type'] == 'deposit'


requests_mock_datadir_missing_one = requests_mock_datadir_factory(ignore_urls=[
    'https://deposit.softwareheritage.org/1/private/666/raw/',
])


def test_deposit_loading_failure_to_retrieve_1_artifact(
        swh_config, requests_mock_datadir_missing_one):
    """Deposit with missing artifact ends up with an uneventful/partial visit

    """
    # private api url form: 'https://deposit.s.o/1/private/hal/666/raw/'
    url = 'some-url-2'
    deposit_id = 666
    loader = DepositLoader(url, deposit_id)

    actual_load_status = loader.load()
    assert actual_load_status['status'] == 'uneventful'
    assert actual_load_status['snapshot_id'] is not None

    stats = get_stats(loader.storage)
    assert {
        'content': 0,
        'directory': 0,
        'origin': 1,
        'origin_visit': 1,
        'person': 0,
        'release': 0,
        'revision': 0,
        'skipped_content': 0,
        'snapshot': 1,
    } == stats

    origin_visit = next(loader.storage.origin_visit_get(url))
    assert origin_visit['status'] == 'partial'
    assert origin_visit['type'] == 'deposit'


def test_revision_metadata_structure(swh_config, requests_mock_datadir):
    # do not care for deposit update query
    requests_mock_datadir.put(re.compile('https'))

    url = 'https://hal-test.archives-ouvertes.fr/some-external-id'
    deposit_id = 666
    loader = DepositLoader(url, deposit_id)

    actual_load_status = loader.load()
    assert actual_load_status['status'] == 'eventful'
    assert actual_load_status['snapshot_id'] is not None

    expected_revision_id = hash_to_bytes(
        '9471c606239bccb1f269564c9ea114e1eeab9eb4')
    revision = list(loader.storage.revision_get([expected_revision_id]))[0]

    assert revision is not None

    check_metadata_paths(revision['metadata'], paths=[
        ('extrinsic.provider', str),
        ('extrinsic.when', str),
        ('extrinsic.raw', dict),
        ('original_artifact', list),
    ])

    for original_artifact in revision['metadata']['original_artifact']:
        check_metadata_paths(original_artifact, paths=[
            ('filename', str),
            ('length', int),
            ('checksums', dict),
        ])


def test_deposit_loading_ok(swh_config, requests_mock_datadir):
    requests_mock_datadir.put(re.compile('https'))  # do not care for put

    url = 'https://hal-test.archives-ouvertes.fr/some-external-id'
    deposit_id = 666
    loader = DepositLoader(url, deposit_id)

    actual_load_status = loader.load()
    expected_snapshot_id = '453f455d0efb69586143cd6b6e5897f9906b53a7'
    assert actual_load_status == {
        'status': 'eventful',
        'snapshot_id': expected_snapshot_id,
    }

    stats = get_stats(loader.storage)
    assert {
        'content': 303,
        'directory': 12,
        'origin': 1,
        'origin_visit': 1,
        'person': 1,
        'release': 0,
        'revision': 1,
        'skipped_content': 0,
        'snapshot': 1,
    } == stats

    origin_visit = next(loader.storage.origin_visit_get(url))
    assert origin_visit['status'] == 'full'
    assert origin_visit['type'] == 'deposit'

    expected_branches = {
        'HEAD': {
            'target': '9471c606239bccb1f269564c9ea114e1eeab9eb4',
            'target_type': 'revision',
        },
    }

    expected_snapshot = {
        'id': expected_snapshot_id,
        'branches': expected_branches,
    }
    check_snapshot(expected_snapshot, storage=loader.storage)

    # check metadata

    tool = {
        "name": "swh-deposit",
        "version": "0.0.1",
        "configuration": {
            "sword_version": "2",
        }
    }

    tool = loader.storage.tool_get(tool)
    assert tool is not None
    assert tool['id'] is not None

    provider = {
        "provider_name": "hal",
        "provider_type": "deposit_client",
        "provider_url": "https://hal-test.archives-ouvertes.fr/",
        "metadata": None,
    }

    provider = loader.storage.metadata_provider_get_by(provider)
    assert provider is not None
    assert provider['id'] is not None

    metadata = list(loader.storage.origin_metadata_get_by(
        url, provider_type='deposit_client'))
    assert metadata is not None
    assert isinstance(metadata, list)
    assert len(metadata) == 1
    metadata0 = metadata[0]

    assert metadata0['provider_id'] == provider['id']
    assert metadata0['provider_type'] == 'deposit_client'
    assert metadata0['tool_id'] == tool['id']



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import os
import shutil
import psutil


def clean_dangling_folders(dirpath, pattern_check, log=None):
    """Clean up potential dangling temporary working folder rooted at
       `dirpath`. Those folders must match a dedicated pattern and not
       belonging to a live pid.

    Args:
        dirpath (str): Path to check for dangling files
        pattern_check (str): A dedicated pattern to check on first
        level directory (e.g `swh.loader.mercurial.`,
        `swh.loader.svn.`)
        log (Logger): Optional logger

    """
    if not os.path.exists(dirpath):
        return
    for filename in os.listdir(dirpath):
        try:
            # pattern: `swh.loader.svn-pid.{noise}`
            if pattern_check not in filename or \
               '-' not in filename:  # silently ignore unknown patterns
                continue
            _, pid = filename.split('-')
            pid = int(pid.split('.')[0])
            if psutil.pid_exists(pid):
                if log:
                    log.debug('PID %s is live, skipping' % pid)
                continue
            path_to_cleanup = os.path.join(dirpath, filename)
            # could be removed concurrently, so check before removal
            if os.path.exists(path_to_cleanup):
                shutil.rmtree(path_to_cleanup)
        except Exception as e:
            if log:
                msg = 'Fail to clean dangling path %s: %s' % (
                    path_to_cleanup, e)
                log.warn(msg)



# Copyright (C) 2015-2016  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

from swh.loader.core.queue import (QueuePerNbElements,
                                   QueuePerNbUniqueElements,
                                   QueuePerSizeAndNbUniqueElements)


class TestQueuePerNbElements(unittest.TestCase):
    def test_simple_queue_behavior(self):
        max_nb_elements = 10
        queue = QueuePerNbElements(max_nb_elements=max_nb_elements)

        elements = [1, 3, 4, 9, 20, 30, 40]
        actual_threshold = queue.add(elements)

        self.assertFalse(actual_threshold, len(elements) > max_nb_elements)

        # pop returns the content and reset the queue
        actual_elements = queue.pop()
        self.assertEqual(actual_elements, elements)
        self.assertEqual(queue.pop(), [])

        # duplicates can be integrated
        new_elements = [1, 1, 3, 4, 9, 20, 30, 40, 12, 14, 2]
        actual_threshold = queue.add(new_elements)

        self.assertTrue(actual_threshold)
        self.assertEqual(queue.pop(), new_elements)

        # reset is destructive too
        queue.add(new_elements)
        queue.reset()

        self.assertEqual(queue.pop(), [])


def to_some_objects(elements, key):
    for elt in elements:
        yield {key: elt}


class TestQueuePerNbUniqueElements(unittest.TestCase):
    def test_queue_with_unique_key_behavior(self):
        max_nb_elements = 5
        queue = QueuePerNbUniqueElements(max_nb_elements=max_nb_elements,
                                         key='id')

        # no duplicates
        elements = list(to_some_objects([1, 1, 3, 4, 9], key='id'))
        actual_threshold = queue.add(elements)

        self.assertFalse(actual_threshold, len(elements) > max_nb_elements)

        # pop returns the content and reset the queue
        actual_elements = queue.pop()
        self.assertEqual(actual_elements,
                         [{'id': 1}, {'id': 3}, {'id': 4}, {'id': 9}])
        self.assertEqual(queue.pop(), [])

        new_elements = list(to_some_objects(
            [1, 3, 4, 9, 20],
            key='id'))
        actual_threshold = queue.add(new_elements)

        self.assertTrue(actual_threshold)

        # reset is destructive too
        queue.add(new_elements)
        queue.reset()

        self.assertEqual(queue.pop(), [])


def to_some_complex_objects(elements, key):
    for elt, size in elements:
        yield {key: elt, 'length': size}


class TestQueuePerSizeAndNbUniqueElements(unittest.TestCase):
    def test_queue_with_unique_key_and_size_behavior(self):
        max_nb_elements = 5
        max_size = 100
        queue = QueuePerSizeAndNbUniqueElements(
            max_nb_elements=max_nb_elements,
            max_size=max_size,
            key='k')

        # size total exceeded, nb elements not reached, still the
        # threshold is deemed reached
        elements = list(to_some_complex_objects([(1, 10),
                                                 (2, 20),
                                                 (3, 30),
                                                 (4, 100)], key='k'))
        actual_threshold = queue.add(elements)

        self.assertTrue(actual_threshold)

        # pop returns the content and reset the queue
        actual_elements = queue.pop()
        self.assertEqual(actual_elements,
                         [{'k': 1, 'length': 10},
                          {'k': 2, 'length': 20},
                          {'k': 3, 'length': 30},
                          {'k': 4, 'length': 100}])
        self.assertEqual(queue.pop(), [])

        # size threshold not reached, nb elements reached, the
        # threshold is considered reached
        new_elements = list(to_some_complex_objects(
            [(1, 10), (3, 5), (4, 2), (9, 1), (20, 0)],
            key='k'))
        actual_threshold = queue.add(new_elements)

        queue.reset()

        self.assertTrue(actual_threshold)

        # nb elements threshold not reached, nor the top number of
        # elements, the threshold is not reached
        new_elements = list(to_some_complex_objects(
            [(1, 10)],
            key='k'))
        actual_threshold = queue.add(new_elements)

        self.assertFalse(actual_threshold)

        # reset is destructive too
        queue.add(new_elements)
        queue.reset()

        self.assertEqual(queue.pop(), [])



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from celery import shared_task

from swh.loader.package.debian import DebianLoader
from swh.loader.package.deposit import DepositLoader
from swh.loader.package.npm import NpmLoader
from swh.loader.package.pypi import PyPILoader
from swh.loader.package.archive import ArchiveLoader


@shared_task(name=__name__ + '.LoadArchive')
def load_archive_files(url=None, artifacts=None, identity_artifact_keys=None):
    """Load archive's artifacts (e.g gnu, etc...)"""
    return ArchiveLoader(url, artifacts,
                         identity_artifact_keys=identity_artifact_keys).load()


@shared_task(name=__name__ + '.LoadDebian')
def load_deb_package(*, url, date, packages):
    """Load Debian package"""
    return DebianLoader(url, date, packages).load()


@shared_task(name=__name__ + '.LoadDeposit')
def load_deposit(*, url, deposit_id):
    """Load Deposit artifacts"""
    return DepositLoader(url, deposit_id).load()


@shared_task(name=__name__ + '.LoadNpm')
def load_npm(*, package_name, package_url, package_metadata_url):
    """Load Npm package"""
    return NpmLoader(package_name, package_url, package_metadata_url).load()


@shared_task(name=__name__ + '.LoadPyPI')
def load_pypi(*, url=None):
    """Load PyPI package"""
    return PyPILoader(url).load()
